import hashlib
import os
import re
from binascii import hexlify, unhexlify
from hive.utils import compat_bytes, compat_chr

import ecdsa

from .base58 import ripemd160, Base58
from .dictionary import words as BrainKeyDictionary


class PasswordKey(object):
    """This class derives a private key given the account name, the
    role and a password. It leverages the technology of Brainkeys
    and allows people to have a secure private key by providing a
    passphrase only.
    """

    def __init__(self, account, password, role="active"):
        self.account = account
        self.role = role
        self.password = password

    def get_private(self):
        """Derive private key from the brain key and the current sequence
        number
        """
        a = compat_bytes(self.account + self.role + self.password, "utf8")
        s = hashlib.sha256(a).digest()
        return PrivateKey(hexlify(s).decode("ascii"))

    def get_public(self):
        return self.get_private().pubkey

    def get_private_key(self):
        return self.get_private()

    def get_public_key(self):
        return self.get_public()


class BrainKey(object):
    """Brainkey implementation similar to the graphene-ui web-wallet.

    :param str brainkey: Brain Key
    :param int sequence: Sequence number for consecutive keys

    Keys in Graphene are derived from a seed brain key which is a string of
    16 words out of a predefined dictionary with 49744 words. It is a
    simple single-chain key derivation scheme that is not compatible with
    BIP44 but easy to use.

    Given the brain key, a private key is derived as::

        privkey = SHA256(SHA512(brainkey + " " + sequence))

    Incrementing the sequence number yields a new key that can be
    regenerated given the brain key.

    """

    def __init__(self, brainkey=None, sequence=0):
        if not brainkey:
            self.brainkey = self.suggest()
        else:
            self.brainkey = self.normalize(brainkey).strip()
        self.sequence = sequence

    def __next__(self):
        """Get the next private key (sequence number increment) for
        iterators
        """
        return self.next_sequence()

    def next_sequence(self):
        """ Increment the sequence number by 1 """
        self.sequence += 1
        return self

    def normalize(self, brainkey):
        """Correct formating with single whitespace syntax and no trailing
        space"""
        return " ".join(re.compile("[\t\n\v\f\r ]+").split(brainkey))

    def get_brainkey(self):
        """ Return brain key of this instance """
        return self.normalize(self.brainkey)

    def get_private(self):
        """Derive private key from the brain key and the current sequence
        number
        """
        encoded = "%s %d" % (self.brainkey, self.sequence)
        a = compat_bytes(encoded, "ascii")

        s = hashlib.sha256(hashlib.sha512(a).digest()).digest()
        return PrivateKey(hexlify(s).decode("ascii"))

    def get_public(self):
        return self.get_private().pubkey

    def get_private_key(self):
        return self.get_private()

    def get_public_key(self):
        return self.get_public()

    def suggest(self):
        """Suggest a new random brain key. Randomness is provided by the
        operating system using ``os.urandom()``.
        """
        word_count = 16
        brainkey = [None] * word_count
        dict_lines = BrainKeyDictionary.split(",")
        assert len(dict_lines) == 49744
        for j in range(0, word_count):
            num = int.from_bytes(os.urandom(2), byteorder="little")
            rndMult = num / 2 ** 16  # returns float between 0..1 (inclusive)
            wIdx = round(len(dict_lines) * rndMult)
            brainkey[j] = dict_lines[wIdx]
        return " ".join(brainkey).upper()


class Address(object):
    """Address class

    This class serves as an address representation for Public Keys.

    :param str address: Base58 encoded address (defaults to ``None``)
    :param str pubkey: Base58 encoded pubkey (defaults to ``None``)
    :param str prefix: Network prefix (defaults to ``GPH``)

    Example::

       Address("GPHFN9r6VYzBK8EKtMewfNbfiGCr56pHDBFi")

    """

    def __init__(self, address=None, pubkey=None, prefix="STM"):
        self.prefix = prefix
        if pubkey is not None:
            self._pubkey = Base58(pubkey, prefix=prefix)
            self._address = None
        elif address is not None:
            self._pubkey = None
            self._address = Base58(address, prefix=prefix)
        else:
            raise Exception(
                "Address has to be initialized by either the "
                + "pubkey or the address."
            )

    def derivesha256address(self):
        """ Derive address using ``RIPEMD160(SHA256(x))`` """
        pkbin = unhexlify(repr(self._pubkey))
        addressbin = ripemd160(hexlify(hashlib.sha256(pkbin).digest()))
        return Base58(hexlify(addressbin).decode("ascii"))

    def derivesha512address(self):
        """ Derive address using ``RIPEMD160(SHA512(x))`` """
        pkbin = unhexlify(repr(self._pubkey))
        addressbin = ripemd160(hexlify(hashlib.sha512(pkbin).digest()))
        return Base58(hexlify(addressbin).decode("ascii"))

    def __repr__(self):
        """Gives the hex representation of the ``GrapheneBase58CheckEncoded``
        Graphene address.
        """
        return repr(self.derivesha512address())

    def __str__(self):
        """Returns the readable Graphene address. This call is equivalent to
        ``format(Address, "GPH")``
        """
        return format(self, self.prefix)

    def __format__(self, _format):
        """May be issued to get valid "MUSE", "PLAY" or any other Graphene compatible
        address with corresponding prefix.
        """
        if self._address is None:
            if _format.lower() == "btc":
                return format(self.derivesha256address(), _format)
            else:
                return format(self.derivesha512address(), _format)
        else:
            return format(self._address, _format)

    def __bytes__(self):
        """ Returns the raw content of the ``Base58CheckEncoded`` address """
        if self._address is None:
            return compat_bytes(self.derivesha512address())
        else:
            return compat_bytes(self._address)


class PublicKey(object):
    """This class deals with Public Keys and inherits ``Address``.

    :param str pk: Base58 encoded public key
    :param str prefix: Network prefix (defaults to ``GPH``)

    Example:::

       PublicKey("GPH6UtYWWs3rkZGV8JA86qrgkG6tyFksgECefKE1MiH4HkLD8PFGL")

    .. note:: By default, graphene-based networks deal with **compressed**
              public keys. If an **uncompressed** key is required, the
              method ``unCompressed`` can be used::

                  PublicKey("xxxxx").unCompressed()

    """

    def __init__(self, pk, prefix="STM"):
        self.prefix = prefix
        self._pk = Base58(pk, prefix=prefix)
        self.address = Address(pubkey=pk, prefix=prefix)
        self.pubkey = self._pk

    def _derive_y_from_x(self, x, is_even):
        """ Derive y point from x point """
        curve = ecdsa.SECP256k1.curve
        # The curve equation over F_p is:
        #   y^2 = x^3 + ax + b
        a, b, p = curve.a(), curve.b(), curve.p()
        alpha = (pow(x, 3, p) + a * x + b) % p
        beta = ecdsa.numbertheory.square_root_mod_prime(alpha, p)
        if (beta % 2) == is_even:
            beta = p - beta
        return beta

    def compressed(self):
        """ Derive compressed public key """
        order = ecdsa.SECP256k1.generator.order()
        p = ecdsa.VerifyingKey.from_string(
            compat_bytes(self), curve=ecdsa.SECP256k1
        ).pubkey.point
        x_str = ecdsa.util.number_to_string(p.x(), order)
        # y_str = ecdsa.util.number_to_string(p.y(), order)
        compressed = hexlify(
            compat_bytes(compat_chr(2 + (p.y() & 1)), "ascii") + x_str
        ).decode("ascii")
        return compressed

    def unCompressed(self):
        """ Derive uncompressed key """
        public_key = repr(self._pk)
        prefix = public_key[0:2]
        if prefix == "04":
            return public_key
        assert prefix == "02" or prefix == "03"
        x = int(public_key[2:], 16)
        y = self._derive_y_from_x(x, (prefix == "02"))
        key = "04" + "%064x" % x + "%064x" % y
        return key

    def point(self):
        """ Return the point for the public key """
        string = unhexlify(self.unCompressed())
        return ecdsa.VerifyingKey.from_string(
            string[1:], curve=ecdsa.SECP256k1
        ).pubkey.point

    def __repr__(self):
        """ Gives the hex representation of the Graphene public key. """
        return repr(self._pk)

    def __str__(self):
        """Returns the readable Graphene public key. This call is equivalent to
        ``format(PublicKey, "GPH")``
        """
        return format(self._pk, self.prefix)

    def __format__(self, _format):
        """Formats the instance of:doc:`Base58 <base58>` according
        to ``_format``"""
        return format(self._pk, _format)

    def __bytes__(self):
        """ Returns the raw public key (has length 33)"""
        return compat_bytes(self._pk)


class PrivateKey(object):
    """Derives the compressed and uncompressed public keys and
    constructs two instances of ``PublicKey``:

    :param str wif: Base58check-encoded wif key
    :param str prefix: Network prefix (defaults to ``GPH``)

    Example:::

        PrivateKey("5HqUkGuo62BfcJU5vNhTXKJRXuUi9QSE6jp8C3uBJ2BVHtB8WSd")

    Compressed vs. Uncompressed:

    * ``PrivateKey("w-i-f").pubkey``:
        Instance of ``PublicKey`` using compressed key.
    * ``PrivateKey("w-i-f").pubkey.address``:
        Instance of ``Address`` using compressed key.
    * ``PrivateKey("w-i-f").uncompressed``:
        Instance of ``PublicKey`` using uncompressed key.
    * ``PrivateKey("w-i-f").uncompressed.address``:
        Instance of ``Address`` using uncompressed key.

    """

    def __init__(self, wif=None, prefix="SHARE"):
        if wif is None:
            import os

            self._wif = Base58(hexlify(os.urandom(32)).decode("ascii"))
        elif isinstance(wif, Base58):
            self._wif = wif
        else:
            self._wif = Base58(wif)
        # compress pubkeys only
        self._pubkeyhex, self._pubkeyuncompressedhex = self.compressedpubkey()
        self.pubkey = PublicKey(self._pubkeyhex, prefix=prefix)
        self.uncompressed = PublicKey(self._pubkeyuncompressedhex, prefix=prefix)
        self.uncompressed.address = Address(
            pubkey=self._pubkeyuncompressedhex, prefix=prefix
        )
        self.address = Address(pubkey=self._pubkeyhex, prefix=prefix)

    def compressedpubkey(self):
        """ Derive uncompressed public key """
        secret = unhexlify(repr(self._wif))
        order = ecdsa.SigningKey.from_string(
            secret, curve=ecdsa.SECP256k1
        ).curve.generator.order()
        p = ecdsa.SigningKey.from_string(
            secret, curve=ecdsa.SECP256k1
        ).verifying_key.pubkey.point
        x_str = ecdsa.util.number_to_string(p.x(), order)
        y_str = ecdsa.util.number_to_string(p.y(), order)
        compressed = hexlify(
            compat_chr(2 + (p.y() & 1)).encode("ascii") + x_str
        ).decode("ascii")
        uncompressed = hexlify(compat_chr(4).encode("ascii") + x_str + y_str).decode(
            "ascii"
        )
        return [compressed, uncompressed]

    def __format__(self, _format):
        """Formats the instance of:doc:`Base58 <base58>` according to
        ``_format``
        """
        return format(self._wif, _format)

    def __repr__(self):
        """ Gives the hex representation of the Graphene private key."""
        return repr(self._wif)

    def __str__(self):
        """Returns the readable (uncompressed wif format) Graphene private key. This
        call is equivalent to ``format(PrivateKey, "WIF")``
        """
        return format(self._wif, "WIF")

    def __bytes__(self):
        """ Returns the raw private key """
        return compat_bytes(self._wif)



import argparse
import json
import logging
import os
import pkg_resources
import pprint
import re
import hive as hve
import sys

from prettytable import PrettyTable
from hivebase.storage import configStorage
from hivebase.account import PrivateKey

from .account import Account
from .amount import Amount
from .block import Block
from .blockchain import Blockchain
from .dex import Dex
from .instance import shared_hived_instance
from .post import Post
from .utils import construct_identifier, strfage
from .witness import Witness

availableConfigurationKeys = [
    "default_account",
    "default_vote_weight",
    "nodes",
]


def legacyentry():
    """
    Piston like cli application.
    This will be re-written as a @click app in the future.
    """
    global args

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="Command line tool to interact with the Hive network",
    )
    """
        Default settings for all tools
    """
    parser.add_argument(
        "--node",
        type=str,
        default=configStorage["node"],
        help='URL for public Hive API (default: "https://api.hive.blog")',
    )

    parser.add_argument(
        "--no-broadcast", "-d", action="store_true", help="Do not broadcast anything"
    )
    parser.add_argument(
        "--no-wallet", "-p", action="store_true", help="Do not load the wallet"
    )
    parser.add_argument(
        "--unsigned",
        "-x",
        action="store_true",
        help="Do not try to sign the transaction",
    )
    parser.add_argument(
        "--expires",
        "-e",
        default=60,
        help="Expiration time in seconds (defaults to 30)",
    )
    parser.add_argument("--verbose", "-v", type=int, default=3, help="Verbosity")
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s {version}".format(
            version=pkg_resources.require("steem")[0].version
        ),
    )

    subparsers = parser.add_subparsers(help="sub-command help")
    """
        Command "set"
    """
    setconfig = subparsers.add_parser("set", help="Set configuration")
    setconfig.add_argument(
        "key", type=str, choices=availableConfigurationKeys, help="Configuration key"
    )
    setconfig.add_argument("value", type=str, help="Configuration value")
    setconfig.set_defaults(command="set")
    """
        Command "config"
    """
    configconfig = subparsers.add_parser("config", help="Show local configuration")
    configconfig.set_defaults(command="config")
    """
        Command "info"
    """
    parser_info = subparsers.add_parser("info", help="Show basic Hive blockchain info")
    parser_info.set_defaults(command="info")
    parser_info.add_argument(
        "objects",
        nargs="*",
        type=str,
        help="General information about the blockchain, a block, an account"
        " name, a post, a public key, ...",
    )
    """
        Command "changewalletpassphrase"
    """
    changepasswordconfig = subparsers.add_parser(
        "changewalletpassphrase", help="Change wallet password"
    )
    changepasswordconfig.set_defaults(command="changewalletpassphrase")
    """
        Command "addkey"
    """
    addkey = subparsers.add_parser("addkey", help="Add a new key to the wallet")
    addkey.add_argument(
        "--unsafe-import-key",
        nargs="*",
        type=str,
        help="private key to import into the wallet (unsafe, unless you "
        + "delete your shell history)",
    )
    addkey.set_defaults(command="addkey")

    parsewif = subparsers.add_parser(
        "parsewif", help="Parse a WIF private key without importing"
    )
    parsewif.add_argument(
        "--unsafe-import-key",
        nargs="*",
        type=str,
        help="WIF key to parse (unsafe, delete your bash history)",
    )
    parsewif.set_defaults(command="parsewif")
    """
        Command "delkey"
    """
    delkey = subparsers.add_parser("delkey", help="Delete keys from the wallet")
    delkey.add_argument(
        "pub", nargs="*", type=str, help="the public key to delete from the wallet"
    )
    delkey.set_defaults(command="delkey")
    """
        Command "getkey"
    """
    getkey = subparsers.add_parser(
        "getkey", help="Dump the privatekey of a pubkey from the wallet"
    )
    getkey.add_argument(
        "pub", type=str, help="the public key for which to show the private key"
    )
    getkey.set_defaults(command="getkey")
    """
        Command "listkeys"
    """
    listkeys = subparsers.add_parser(
        "listkeys", help="List available keys in your wallet"
    )
    listkeys.set_defaults(command="listkeys")
    """
        Command "listaccounts"
    """
    listaccounts = subparsers.add_parser(
        "listaccounts", help="List available accounts in your wallet"
    )
    listaccounts.set_defaults(command="listaccounts")
    """
        Command "upvote"
    """
    parser_upvote = subparsers.add_parser("upvote", help="Upvote a post")
    parser_upvote.set_defaults(command="upvote")
    parser_upvote.add_argument(
        "post",
        type=str,
        help="author/permlink-identifier of the post to upvote "
        + "to (e.g. xeroc/python-steem-0-1)",
    )
    parser_upvote.add_argument(
        "--account",
        type=str,
        required=False,
        default=configStorage["default_account"],
        help="The voter account name",
    )
    parser_upvote.add_argument(
        "--weight",
        type=float,
        default=configStorage["default_vote_weight"],
        required=False,
        help="Actual weight (from 0.1 to 100.0)",
    )
    """
        Command "downvote"
    """
    parser_downvote = subparsers.add_parser("downvote", help="Downvote a post")
    parser_downvote.set_defaults(command="downvote")
    parser_downvote.add_argument(
        "--account",
        type=str,
        default=configStorage["default_account"],
        help="The voter account name",
    )
    parser_downvote.add_argument(
        "post",
        type=str,
        help="author/permlink-identifier of the post to downvote "
        + "to (e.g. xeroc/python-steem-0-1)",
    )
    parser_downvote.add_argument(
        "--weight",
        type=float,
        default=configStorage["default_vote_weight"],
        required=False,
        help="Actual weight (from 0.1 to 100.0)",
    )
    """
        Command "transfer"
    """
    parser_transfer = subparsers.add_parser("transfer", help="Transfer HIVE")
    parser_transfer.set_defaults(command="transfer")
    parser_transfer.add_argument("to", type=str, help="Recipient")
    parser_transfer.add_argument("amount", type=float, help="Amount to transfer")
    parser_transfer.add_argument(
        "asset",
        type=str,
        choices=["HIVE", "HBD"],
        help="Asset to transfer (i.e. HIVE or HBD)",
    )
    parser_transfer.add_argument(
        "memo", type=str, nargs="?", default="", help="Optional memo"
    )
    parser_transfer.add_argument(
        "--account",
        type=str,
        required=False,
        default=configStorage["default_account"],
        help="Transfer from this account",
    )
    """
        Command "powerup"
    """
    parser_powerup = subparsers.add_parser(
        "powerup", help="Power up (vest HIVE as HIVE POWER)"
    )
    parser_powerup.set_defaults(command="powerup")
    parser_powerup.add_argument("amount", type=str, help="Amount of VESTS to powerup")
    parser_powerup.add_argument(
        "--account",
        type=str,
        required=False,
        default=configStorage["default_account"],
        help="Powerup from this account",
    )
    parser_powerup.add_argument(
        "--to", type=str, required=False, default=None, help="Powerup this account"
    )
    """
        Command "powerdown"
    """
    parser_powerdown = subparsers.add_parser(
        "powerdown", help="Power down (start withdrawing HIVE from HIVE POWER)"
    )
    parser_powerdown.set_defaults(command="powerdown")
    parser_powerdown.add_argument(
        "amount", type=str, help="Amount of VESTS to powerdown"
    )
    parser_powerdown.add_argument(
        "--account",
        type=str,
        required=False,
        default=configStorage["default_account"],
        help="powerdown from this account",
    )
    """
        Command "powerdownroute"
    """
    parser_powerdownroute = subparsers.add_parser(
        "powerdownroute", help="Setup a powerdown route"
    )
    parser_powerdownroute.set_defaults(command="powerdownroute")
    parser_powerdownroute.add_argument(
        "to",
        type=str,
        default=configStorage["default_account"],
        help="The account receiving either VESTS/HivePower or HIVE.",
    )
    parser_powerdownroute.add_argument(
        "--percentage",
        type=float,
        default=100,
        help='The percent of the withdraw to go to the "to" account',
    )
    parser_powerdownroute.add_argument(
        "--account",
        type=str,
        default=configStorage["default_account"],
        help="The account which is powering down",
    )
    parser_powerdownroute.add_argument(
        "--auto_vest",
        action="store_true",
        help=(
            "Set to true if the from account should receive the VESTS as"
            "VESTS, or false if it should receive them as HIVE."
        ),
    )
    """
        Command "convert"
    """
    parser_convert = subparsers.add_parser(
        "convert", help="Convert HBD to HIVE (takes 3.5 days to settle)"
    )
    parser_convert.set_defaults(command="convert")
    parser_convert.add_argument("amount", type=float, help="Amount of HBD to convert")
    parser_convert.add_argument(
        "--account",
        type=str,
        required=False,
        default=configStorage["default_account"],
        help="Convert from this account",
    )
    """
        Command "balance"
    """
    parser_balance = subparsers.add_parser(
        "balance", help="Show the balance of one more more accounts"
    )
    parser_balance.set_defaults(command="balance")
    parser_balance.add_argument(
        "account",
        type=str,
        nargs="*",
        default=configStorage["default_account"],
        help="balance of these account (multiple accounts allowed)",
    )
    """
        Command "interest"
    """
    interest = subparsers.add_parser(
        "interest", help="Get information about interest payment"
    )
    interest.set_defaults(command="interest")
    interest.add_argument(
        "account",
        type=str,
        nargs="*",
        default=configStorage["default_account"],
        help="Inspect these accounts",
    )
    """
        Command "permissions"
    """
    parser_permissions = subparsers.add_parser(
        "permissions", help="Show permissions of an account"
    )
    parser_permissions.set_defaults(command="permissions")
    parser_permissions.add_argument(
        "account",
        type=str,
        nargs="?",
        default=configStorage["default_account"],
        help="Account to show permissions for",
    )
    """
        Command "allow"
    """
    parser_allow = subparsers.add_parser(
        "allow", help="Allow an account/key to interact with your account"
    )
    parser_allow.set_defaults(command="allow")
    parser_allow.add_argument(
        "--account",
        type=str,
        nargs="?",
        default=configStorage["default_account"],
        help="The account to allow action for",
    )
    parser_allow.add_argument(
        "foreign_account",
        type=str,
        nargs="?",
        help="The account or key that will be allowed to interact with "
        + "your account",
    )
    parser_allow.add_argument(
        "--permission",
        type=str,
        default="posting",
        choices=["owner", "posting", "active"],
        help='The permission to grant (defaults to "posting")',
    )
    parser_allow.add_argument(
        "--weight",
        type=int,
        default=None,
        help=(
            "The weight to use instead of the (full) threshold. "
            "If the weight is smaller than the threshold, "
            "additional signatures are required"
        ),
    )
    parser_allow.add_argument(
        "--threshold",
        type=int,
        default=None,
        help=(
            "The permission's threshold that needs to be reached "
            "by signatures to be able to interact"
        ),
    )
    """
        Command "disallow"
    """
    parser_disallow = subparsers.add_parser(
        "disallow", help="Remove allowance an account/key to interact with your account"
    )
    parser_disallow.set_defaults(command="disallow")
    parser_disallow.add_argument(
        "--account",
        type=str,
        nargs="?",
        default=configStorage["default_account"],
        help="The account to disallow action for",
    )
    parser_disallow.add_argument(
        "foreign_account",
        type=str,
        help="The account or key whose allowance to interact as your "
        + "account will be removed",
    )
    parser_disallow.add_argument(
        "--permission",
        type=str,
        default="posting",
        choices=["owner", "posting", "active"],
        help='The permission to remove (defaults to "posting")',
    )
    parser_disallow.add_argument(
        "--threshold",
        type=int,
        default=None,
        help=(
            "The permission's threshold that needs to be reached "
            "by signatures to be able to interact"
        ),
    )
    """
        Command "newaccount"
    """
    parser_newaccount = subparsers.add_parser("newaccount", help="Create a new account")
    parser_newaccount.set_defaults(command="newaccount")
    parser_newaccount.add_argument("accountname", type=str, help="New account name")
    parser_newaccount.add_argument(
        "--account",
        type=str,
        required=False,
        default=configStorage["default_account"],
        help="Account that pays the fee",
    )
    parser_newaccount.add_argument(
        "--fee",
        type=str,
        required=False,
        default="0 HIVE",
        help="Base Fee to pay. Delegate the rest.",
    )
    """
        Command "importaccount"
    """
    parser_importaccount = subparsers.add_parser(
        "importaccount", help="Import an account using a passphrase"
    )
    parser_importaccount.set_defaults(command="importaccount")
    parser_importaccount.add_argument("account", type=str, help="Account name")
    parser_importaccount.add_argument(
        "--roles",
        type=str,
        nargs="*",
        default=["active", "posting", "memo"],  # no owner
        help="Import specified keys (owner, active, posting, memo)",
    )
    """
        Command "updateMemoKey"
    """
    parser_updateMemoKey = subparsers.add_parser(
        "updatememokey", help="Update an account's memo key"
    )
    parser_updateMemoKey.set_defaults(command="updatememokey")
    parser_updateMemoKey.add_argument(
        "--account",
        type=str,
        nargs="?",
        default=configStorage["default_account"],
        help="The account to updateMemoKey action for",
    )
    parser_updateMemoKey.add_argument(
        "--key", type=str, default=None, help="The new memo key"
    )
    """
        Command "approvewitness"
    """
    parser_approvewitness = subparsers.add_parser(
        "approvewitness", help="Approve a witnesses"
    )
    parser_approvewitness.set_defaults(command="approvewitness")
    parser_approvewitness.add_argument("witness", type=str, help="Witness to approve")
    parser_approvewitness.add_argument(
        "--account",
        type=str,
        required=False,
        default=configStorage["default_account"],
        help="Your account",
    )
    """
        Command "disapprovewitness"
    """
    parser_disapprovewitness = subparsers.add_parser(
        "disapprovewitness", help="Disapprove a witnesses"
    )
    parser_disapprovewitness.set_defaults(command="disapprovewitness")
    parser_disapprovewitness.add_argument(
        "witness", type=str, help="Witness to disapprove"
    )
    parser_disapprovewitness.add_argument(
        "--account",
        type=str,
        required=False,
        default=configStorage["default_account"],
        help="Your account",
    )
    """
        Command "sign"
    """
    parser_sign = subparsers.add_parser(
        "sign", help="Sign a provided transaction with available and required keys"
    )
    parser_sign.set_defaults(command="sign")
    parser_sign.add_argument(
        "--file",
        type=str,
        required=False,
        help='Load transaction from file. If "-", read from '
        + 'stdin (defaults to "-")',
    )
    """
        Command "broadcast"
    """
    parser_broadcast = subparsers.add_parser(
        "broadcast", help="broadcast a signed transaction"
    )
    parser_broadcast.set_defaults(command="broadcast")
    parser_broadcast.add_argument(
        "--file",
        type=str,
        required=False,
        help='Load transaction from file. If "-", read from '
        + 'stdin (defaults to "-")',
    )
    """
        Command "orderbook"
    """
    orderbook = subparsers.add_parser(
        "orderbook", help="Obtain orderbook of the internal market"
    )
    orderbook.set_defaults(command="orderbook")
    orderbook.add_argument(
        "--chart", action="store_true", help="Enable charting (requires matplotlib)"
    )
    """
        Command "buy"
    """
    parser_buy = subparsers.add_parser(
        "buy", help="Buy HIVE or HBD from the internal market"
    )
    parser_buy.set_defaults(command="buy")
    parser_buy.add_argument("amount", type=float, help="Amount to buy")
    parser_buy.add_argument(
        "asset",
        type=str,
        choices=["HIVE", "HBD"],
        help="Asset to buy (i.e. HIVE or HBD)",
    )
    parser_buy.add_argument(
        "price", type=float, help="Limit buy price denoted in (HBD per HIVE)"
    )
    parser_buy.add_argument(
        "--account",
        type=str,
        required=False,
        default=configStorage["default_account"],
        help='Buy with this account (defaults to "default_account")',
    )
    """
        Command "sell"
    """
    parser_sell = subparsers.add_parser(
        "sell", help="Sell HIVE or HBD from the internal market"
    )
    parser_sell.set_defaults(command="sell")
    parser_sell.add_argument("amount", type=float, help="Amount to sell")
    parser_sell.add_argument(
        "asset",
        type=str,
        choices=["HIVE", "HBD"],
        help="Asset to sell (i.e. HIVE or HBD)",
    )
    parser_sell.add_argument(
        "price", type=float, help="Limit sell price denoted in (HBD per HIVE)"
    )
    parser_sell.add_argument(
        "--account",
        type=str,
        required=False,
        default=configStorage["default_account"],
        help='Sell from this account (defaults to "default_account")',
    )
    """
        Command "cancel"
    """
    parser_cancel = subparsers.add_parser(
        "cancel", help="Cancel order in the internal market"
    )
    parser_cancel.set_defaults(command="cancel")
    parser_cancel.add_argument("orderid", type=int, help="Orderid")
    parser_cancel.add_argument(
        "--account",
        type=str,
        required=False,
        default=configStorage["default_account"],
        help='Cancel from this account (defaults to "default_account")',
    )
    """
        Command "reblog"
    """
    parser_reblog = subparsers.add_parser("reblog", help="Reblog an existing post")
    parser_reblog.set_defaults(command="reblog")
    parser_reblog.add_argument(
        "identifier", type=str, help="author/permlink-identifier of the post to reblog"
    )
    parser_reblog.add_argument(
        "--account",
        type=str,
        required=False,
        default=configStorage["default_account"],
        help="Reblog as this user (requires to have the "
        + "key installed in the wallet)",
    )
    """
        Command "follow"
    """
    parser_follow = subparsers.add_parser("follow", help="Follow another account")
    parser_follow.set_defaults(command="follow")
    parser_follow.add_argument("follow", type=str, help="Account to follow")
    parser_follow.add_argument(
        "--account",
        type=str,
        required=False,
        default=configStorage["default_account"],
        help="Follow from this account",
    )
    parser_follow.add_argument(
        "--what",
        type=str,
        required=False,
        nargs="*",
        default=["blog"],
        help='Follow these objects (defaults to "blog")',
    )
    """
        Command "unfollow"
    """
    parser_unfollow = subparsers.add_parser("unfollow", help="unfollow another account")
    parser_unfollow.set_defaults(command="unfollow")
    parser_unfollow.add_argument("unfollow", type=str, help="Account to unfollow")
    parser_unfollow.add_argument(
        "--account",
        type=str,
        required=False,
        default=configStorage["default_account"],
        help="Unfollow from this account",
    )
    parser_unfollow.add_argument(
        "--what",
        type=str,
        required=False,
        nargs="*",
        default=[],
        help='Unfollow these objects (defaults to "blog")',
    )
    """
        Command "setprofile"
    """
    parser_setprofile = subparsers.add_parser(
        "setprofile", help="Set a variable in an account's profile"
    )
    parser_setprofile.set_defaults(command="setprofile")
    parser_setprofile.add_argument(
        "--account",
        type=str,
        required=False,
        default=configStorage["default_account"],
        help="setprofile as this user (requires to have the key "
        + "installed in the wallet)",
    )
    parser_setprofile_a = parser_setprofile.add_argument_group("Multiple keys at once")
    parser_setprofile_a.add_argument(
        "--pair", type=str, nargs="*", help='"Key=Value" pairs'
    )
    parser_setprofile_b = parser_setprofile.add_argument_group("Just a single key")
    parser_setprofile_b.add_argument(
        "variable", type=str, nargs="?", help="Variable to set"
    )
    parser_setprofile_b.add_argument("value", type=str, nargs="?", help="Value to set")
    """
        Command "delprofile"
    """
    parser_delprofile = subparsers.add_parser(
        "delprofile", help="Set a variable in an account's profile"
    )
    parser_delprofile.set_defaults(command="delprofile")
    parser_delprofile.add_argument(
        "--account",
        type=str,
        required=False,
        default=configStorage["default_account"],
        help="delprofile as this user (requires to have the "
        + "key installed in the wallet)",
    )
    parser_delprofile.add_argument(
        "variable", type=str, nargs="*", help="Variable to set"
    )
    """
        Command "witnessupdate"
    """
    parser_witnessprops = subparsers.add_parser(
        "witnessupdate", help="Change witness properties"
    )
    parser_witnessprops.set_defaults(command="witnessupdate")
    parser_witnessprops.add_argument(
        "--witness",
        type=str,
        default=configStorage["default_account"],
        help="Witness name",
    )
    parser_witnessprops.add_argument(
        "--maximum_block_size", type=float, required=False, help="Max block size"
    )
    parser_witnessprops.add_argument(
        "--account_creation_fee",
        type=float,
        required=False,
        help="Account creation fee",
    )
    parser_witnessprops.add_argument(
        "--hbd_interest_rate",
        type=float,
        required=False,
        help="HBD interest rate in percent",
    )
    parser_witnessprops.add_argument(
        "--url", type=str, required=False, help="Witness URL"
    )
    parser_witnessprops.add_argument(
        "--signing_key", type=str, required=False, help="Signing Key"
    )
    """
        Command "witnesscreate"
    """
    parser_witnesscreate = subparsers.add_parser(
        "witnesscreate", help="Create a witness"
    )
    parser_witnesscreate.set_defaults(command="witnesscreate")
    parser_witnesscreate.add_argument("witness", type=str, help="Witness name")
    parser_witnesscreate.add_argument("signing_key", type=str, help="Signing Key")
    parser_witnesscreate.add_argument(
        "--maximum_block_size", type=float, default="65536", help="Max block size"
    )
    parser_witnesscreate.add_argument(
        "--account_creation_fee", type=float, default=30, help="Account creation fee"
    )
    parser_witnesscreate.add_argument(
        "--hbd_interest_rate",
        type=float,
        default=0.0,
        help="HBD interest rate in percent",
    )
    parser_witnesscreate.add_argument("--url", type=str, default="", help="Witness URL")
    """
        Parse Arguments
    """
    args = parser.parse_args()

    # Logging
    log = logging.getLogger(__name__)
    verbosity = ["critical", "error", "warn", "info", "debug"][
        int(min(args.verbose, 4))
    ]
    log.setLevel(getattr(logging, verbosity.upper()))
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    ch = logging.StreamHandler()
    ch.setLevel(getattr(logging, verbosity.upper()))
    ch.setFormatter(formatter)
    log.addHandler(ch)

    # GrapheneAPI logging
    if args.verbose > 4:
        verbosity = ["critical", "error", "warn", "info", "debug"][
            int(min((args.verbose - 4), 4))
        ]
        gphlog = logging.getLogger("graphenebase")
        gphlog.setLevel(getattr(logging, verbosity.upper()))
        gphlog.addHandler(ch)
    if args.verbose > 8:
        verbosity = ["critical", "error", "warn", "info", "debug"][
            int(min((args.verbose - 8), 4))
        ]
        gphlog = logging.getLogger("grapheneapi")
        gphlog.setLevel(getattr(logging, verbosity.upper()))
        gphlog.addHandler(ch)

    if not hasattr(args, "command"):
        parser.print_help()
        sys.exit(2)

    # initialize Hive instance
    options = {"node": args.node, "unsigned": args.unsigned, "expires": args.expires}
    if args.command == "sign":
        options.update({"offline": True})
    if args.no_wallet:
        options.update({"wif": []})

    hive = hve.Hive(no_broadcast=args.no_broadcast, **options)

    if args.command == "set":
        # TODO: Evaluate this line with cli refactor.
        if args.key in ["default_account"] and args.value[0] == "@":
            args.value = args.value[1:]
        configStorage[args.key] = args.value

    elif args.command == "config":
        t = PrettyTable(["Key", "Value"])
        t.align = "l"
        for key in configStorage:
            # hide internal config data
            if key in availableConfigurationKeys:
                t.add_row([key, configStorage[key]])
        print(t)

    elif args.command == "info":
        if not args.objects:
            t = PrettyTable(["Key", "Value"])
            t.align = "l"
            blockchain = Blockchain(mode="head")
            info = blockchain.info()

            # temporarily allow both steem and hive symbols until after HF24
            if "total_vesting_fund_steem" in info:
                info["total_vesting_fund_hive"] = info["total_vesting_fund_steem"]

            median_price = hive.get_current_median_history_price()
            hive_per_mvest = Amount(info["total_vesting_fund_hive"]).amount / (
                Amount(info["total_vesting_shares"]).amount / 1e6
            )
            price = (
                Amount(median_price["base"]).amount
                / Amount(median_price["quote"]).amount
            )
            for key in info:
                t.add_row([key, info[key]])
            t.add_row(["hive per mvest", hive_per_mvest])
            t.add_row(["internal price", price])
            print(t.get_string(sortby="Key"))

        for obj in args.objects:
            # Block
            if re.match("^[0-9]*$", obj):
                block = Block(obj)
                if block:
                    t = PrettyTable(["Key", "Value"])
                    t.align = "l"
                    for key in sorted(block):
                        value = block[key]
                        if key == "transactions":
                            value = json.dumps(value, indent=4)
                        t.add_row([key, value])
                    print(t)
                else:
                    print("Block number %s unknown" % obj)
            # Account name
            elif re.match("^[a-zA-Z0-9\-\._]{2,16}$", obj):
                from math import log10

                account = Account(obj)
                t = PrettyTable(["Key", "Value"])
                t.align = "l"
                for key in sorted(account):
                    value = account[key]
                    if key == "json_metadata":
                        value = json.dumps(json.loads(value or "{}"), indent=4)
                    if key in ["posting", "witness_votes", "active", "owner"]:
                        value = json.dumps(value, indent=4)
                    if key == "reputation" and int(value) > 0:
                        value = int(value)
                        rep = (
                            max(log10(value) - 9, 0) * 9 + 25
                            if value > 0
                            else max(log10(-value) - 9, 0) * -9 + 25
                        )
                        value = "{:.2f} ({:d})".format(rep, value)
                    t.add_row([key, value])
                print(t)

                # witness available?
                try:
                    witness = Witness(obj)
                    t = PrettyTable(["Key", "Value"])
                    t.align = "l"
                    for key in sorted(witness):
                        value = witness[key]
                        if key in ["props", "hbd_exchange_rate"]:
                            value = json.dumps(value, indent=4)
                        t.add_row([key, value])
                    print(t)
                except:  # noqa FIXME(sneak)
                    pass
            # Public Key
            elif re.match("^STM.{48,55}$", obj):
                account = hive.commit.wallet.getAccountFromPublicKey(obj)
                if account:
                    t = PrettyTable(["Account"])
                    t.align = "l"
                    t.add_row([account])
                    print(t)
                else:
                    print("Public Key not known" % obj)
            # Post identifier
            elif re.match(".*@.{3,16}/.*$", obj):
                post = Post(obj)
                if post:
                    t = PrettyTable(["Key", "Value"])
                    t.align = "l"
                    for key in sorted(post):
                        value = post[key]
                        if key in ["tags", "json_metadata", "active_votes"]:
                            value = json.dumps(value, indent=4)
                        t.add_row([key, value])
                    print(t)
                else:
                    print("Post now known" % obj)
            else:
                print("Couldn't identify object to read")

    elif args.command == "changewalletpassphrase":
        hive.commit.wallet.changeUserPassphrase()

    elif args.command == "addkey":
        if args.unsafe_import_key:
            for key in args.unsafe_import_key:
                try:
                    hive.commit.wallet.addPrivateKey(key)
                except Exception as e:
                    print(str(e))
        else:
            import getpass

            while True:
                wifkey = getpass.getpass("Private Key (wif) [Enter to quit]:")
                if not wifkey:
                    break
                try:
                    hive.commit.wallet.addPrivateKey(wifkey)
                except Exception as e:
                    print(str(e))
                    continue

                installed_keys = hive.commit.wallet.getPublicKeys()
                if len(installed_keys) == 1:
                    name = hive.commit.wallet.getAccountFromPublicKey(installed_keys[0])
                    print("=" * 30)
                    print("Would you like to make %s a default user?" % name)
                    print()
                    print("You can set it with with:")
                    print("    hivepy set default_account <account>")
                    print("=" * 30)

    elif args.command == "delkey":
        if confirm(
            "Are you sure you want to delete keys from your wallet?\n"
            "This step is IRREVERSIBLE! If you don't have a backup, "
            "You may lose access to your account!"
        ):
            for pub in args.pub:
                hive.commit.wallet.removePrivateKeyFromPublicKey(pub)

    elif args.command == "parsewif":
        if args.unsafe_import_key:
            for key in args.unsafe_import_key:
                try:
                    print(PrivateKey(key).pubkey)
                except Exception as e:
                    print(str(e))
        else:
            import getpass

            while True:
                wifkey = getpass.getpass("Private Key (wif) [Enter to quit:")
                if not wifkey:
                    break
                try:
                    print(PrivateKey(wifkey).pubkey)
                except Exception as e:
                    print(str(e))
                    continue
    elif args.command == "getkey":
        print(hive.commit.wallet.getPrivateKeyForPublicKey(args.pub))

    elif args.command == "listkeys":
        t = PrettyTable(["Available Key"])
        t.align = "l"
        for key in hive.commit.wallet.getPublicKeys():
            t.add_row([key])
        print(t)

    elif args.command == "listaccounts":
        t = PrettyTable(["Name", "Type", "Available Key"])
        t.align = "l"
        for account in hive.commit.wallet.getAccounts():
            t.add_row(
                [account["name"] or "n/a", account["type"] or "n/a", account["pubkey"]]
            )
        print(t)

    elif args.command == "upvote" or args.command == "downvote":
        post = Post(args.post)
        if args.command == "downvote":
            weight = -float(args.weight)
        else:
            weight = +float(args.weight)
        if not args.account:
            print("Not voter provided!")
            return
        print_json(post.vote(weight, voter=args.account))

    elif args.command == "transfer":
        print_json(
            hive.commit.transfer(
                args.to, args.amount, args.asset, memo=args.memo, account=args.account
            )
        )

    elif args.command == "powerup":
        print_json(
            hive.commit.transfer_to_vesting(
                args.amount, account=args.account, to=args.to
            )
        )

    elif args.command == "powerdown":
        print_json(hive.commit.withdraw_vesting(args.amount, account=args.account,))

    elif args.command == "convert":
        print_json(hive.commit.convert(args.amount, account=args.account,))

    elif args.command == "powerdownroute":
        print_json(
            hive.commit.set_withdraw_vesting_route(
                args.to,
                percentage=args.percentage,
                account=args.account,
                auto_vest=args.auto_vest,
            )
        )

    elif args.command == "balance":
        if args.account and isinstance(args.account, list):
            for account in args.account:
                a = Account(account, hived_instance=hive)

                print("\n%s" % a.name)
                t = PrettyTable(["Account", "HIVE", "HBD", "VESTS"])
                t.align = "r"
                t.add_row(
                    [
                        "Available",
                        a.balances["available"]["HIVE"],
                        a.balances["available"]["HBD"],
                        a.balances["available"]["VESTS"],
                    ]
                )
                t.add_row(
                    [
                        "Rewards",
                        a.balances["rewards"]["HIVE"],
                        a.balances["rewards"]["HBD"],
                        a.balances["rewards"]["VESTS"],
                    ]
                )
                t.add_row(
                    [
                        "TOTAL",
                        a.balances["total"]["HIVE"],
                        a.balances["total"]["HBD"],
                        a.balances["total"]["VESTS"],
                    ]
                )
                print(t)
        else:
            print("Please specify an account: hivepy balance <account>")

    elif args.command == "interest":
        t = PrettyTable(
            [
                "Account",
                "Last Interest Payment",
                "Next Payment",
                "Interest rate",
                "Interest",
            ]
        )
        t.align = "r"
        if isinstance(args.account, str):
            args.account = [args.account]
        for a in args.account:
            i = hive.commit.interest(a)

            t.add_row(
                [
                    a,
                    i["last_payment"],
                    "in %s" % strfage(i["next_payment_duration"]),
                    "%.1f%%" % i["interest_rate"],
                    "%.3f %s" % (i["interest"], "HBD"),
                ]
            )
        print(t)

    elif args.command == "permissions":
        account = Account(args.account)
        print_permissions(account)

    elif args.command == "allow":
        if not args.foreign_account:
            from hivebase.account import PasswordKey

            pwd = get_terminal(text="Password for Key Derivation: ", confirm=True)
            args.foreign_account = format(
                PasswordKey(args.account, pwd, args.permission).get_public(), "STM"
            )
        print_json(
            hive.commit.allow(
                args.foreign_account,
                weight=args.weight,
                account=args.account,
                permission=args.permission,
                threshold=args.threshold,
            )
        )

    elif args.command == "disallow":
        print_json(
            hive.commit.disallow(
                args.foreign_account,
                account=args.account,
                permission=args.permission,
                threshold=args.threshold,
            )
        )

    elif args.command == "updatememokey":
        if not args.key:
            # Loop until both match
            from hivebase.account import PasswordKey

            pw = get_terminal(
                text="Password for Memo Key: ", confirm=True, allowedempty=False
            )
            memo_key = PasswordKey(args.account, pw, "memo")
            args.key = format(memo_key.get_public_key(), "STM")
            memo_privkey = memo_key.get_private_key()
            # Add the key to the wallet
            if not args.no_broadcast:
                hive.commit.wallet.addPrivateKey(memo_privkey)
        print_json(hive.commit.update_memo_key(args.key, account=args.account))

    elif args.command == "newaccount":
        import getpass

        while True:
            pw = getpass.getpass("New Account Passphrase: ")
            if not pw:
                print("You cannot chosen an empty password!")
                continue
            else:
                pwck = getpass.getpass("Confirm New Account Passphrase: ")
                if pw == pwck:
                    break
                else:
                    print("Given Passphrases do not match!")
        print_json(
            hive.commit.create_account(
                args.accountname,
                creator=args.account,
                password=pw,
                delegation_fee_hive=args.fee,
            )
        )

    elif args.command == "importaccount":
        from hivebase.account import PasswordKey
        import getpass

        password = getpass.getpass("Account Passphrase: ")
        account = Account(args.account)
        imported = False

        if "owner" in args.roles:
            owner_key = PasswordKey(args.account, password, role="owner")
            owner_pubkey = format(owner_key.get_public_key(), "STM")
            if owner_pubkey in [x[0] for x in account["owner"]["key_auths"]]:
                print("Importing owner key!")
                owner_privkey = owner_key.get_private_key()
                hive.commit.wallet.addPrivateKey(owner_privkey)
                imported = True

        if "active" in args.roles:
            active_key = PasswordKey(args.account, password, role="active")
            active_pubkey = format(active_key.get_public_key(), "STM")
            if active_pubkey in [x[0] for x in account["active"]["key_auths"]]:
                print("Importing active key!")
                active_privkey = active_key.get_private_key()
                hive.commit.wallet.addPrivateKey(active_privkey)
                imported = True

        if "posting" in args.roles:
            posting_key = PasswordKey(args.account, password, role="posting")
            posting_pubkey = format(posting_key.get_public_key(), "STM")
            if posting_pubkey in [x[0] for x in account["posting"]["key_auths"]]:
                print("Importing posting key!")
                posting_privkey = posting_key.get_private_key()
                hive.commit.wallet.addPrivateKey(posting_privkey)
                imported = True

        if "memo" in args.roles:
            memo_key = PasswordKey(args.account, password, role="memo")
            memo_pubkey = format(memo_key.get_public_key(), "STM")
            if memo_pubkey == account["memo_key"]:
                print("Importing memo key!")
                memo_privkey = memo_key.get_private_key()
                hive.commit.wallet.addPrivateKey(memo_privkey)
                imported = True

        if not imported:
            print("No matching key(s) found. Password correct?")

    elif args.command == "sign":
        if args.file and args.file != "-":
            if not os.path.isfile(args.file):
                raise Exception("File %s does not exist!" % args.file)
            with open(args.file) as fp:
                tx = fp.read()
        else:
            tx = sys.stdin.read()
        tx = eval(tx)
        print_json(hive.commit.sign(tx))

    elif args.command == "broadcast":
        if args.file and args.file != "-":
            if not os.path.isfile(args.file):
                raise Exception("File %s does not exist!" % args.file)
            with open(args.file) as fp:
                tx = fp.read()
        else:
            tx = sys.stdin.read()
        tx = eval(tx)
        hive.commit.broadcast(tx)

    elif args.command == "buy":
        if args.asset == "HBD":
            price = 1.0 / args.price
        else:
            price = args.price
        dex = Dex(hive)
        print_json(dex.buy(args.amount, args.asset, price, account=args.account))

    elif args.command == "sell":
        if args.asset == "HBD":
            price = 1.0 / args.price
        else:
            price = args.price
        dex = Dex(hive)
        print_json(dex.sell(args.amount, args.asset, price, account=args.account))

    elif args.command == "cancel":
        dex = Dex(hive)
        print_json(dex.cancel(args.orderid))

    elif args.command == "approvewitness":
        print_json(hive.commit.approve_witness(args.witness, account=args.account))

    elif args.command == "disapprovewitness":
        print_json(hive.commit.disapprove_witness(args.witness, account=args.account))

    elif args.command == "reblog":
        print_json(hive.commit.reblog(args.identifier, account=args.account))

    elif args.command == "follow":
        print_json(
            hive.commit.follow(args.follow, what=args.what, account=args.account)
        )

    elif args.command == "unfollow":
        print_json(
            hive.commit.unfollow(args.unfollow, what=args.what, account=args.account)
        )

    elif args.command == "setprofile":
        from .profile import Profile

        keys = []
        values = []
        if args.pair:
            for pair in args.pair:
                key, value = pair.split("=")
                keys.append(key)
                values.append(value)
        if args.variable and args.value:
            keys.append(args.variable)
            values.append(args.value)

        profile = Profile(keys, values)

        account = Account(args.account)
        account["json_metadata"] = Profile(
            account["json_metadata"] if account["json_metadata"] else {}
        )
        account["json_metadata"].update(profile)

        print_json(
            hive.commit.update_account_profile(
                account["json_metadata"], account=args.account
            )
        )

    elif args.command == "delprofile":
        from .profile import Profile

        account = Account(args.account)
        account["json_metadata"] = Profile(account["json_metadata"])

        for var in args.variable:
            account["json_metadata"].remove(var)

        print_json(
            hive.commit.update_account_profile(
                account["json_metadata"], account=args.account
            )
        )

    elif args.command == "witnessupdate":

        witness = Witness(args.witness)
        props = witness["props"]
        if args.account_creation_fee:
            props["account_creation_fee"] = str(
                Amount("%f HIVE" % args.account_creation_fee)
            )
        if args.maximum_block_size:
            props["maximum_block_size"] = args.maximum_block_size
        if args.hbd_interest_rate:
            props["hbd_interest_rate"] = int(args.hbd_interest_rate * 100)

        print_json(
            hive.commit.witness_update(
                args.signing_key or witness["signing_key"],
                args.url or witness["url"],
                props,
                account=args.witness,
            )
        )

    elif args.command == "witnesscreate":
        props = {
            "account_creation_fee": str(Amount("%f HIVE" % args.account_creation_fee)),
            "maximum_block_size": args.maximum_block_size,
            "hbd_interest_rate": int(args.hbd_interest_rate * 100),
        }
        print_json(
            hive.commit.witness_update(
                args.signing_key, args.url, props, account=args.witness
            )
        )

    else:
        print("No valid command given")


def confirm(question, default="yes"):
    """ Confirmation dialog that requires *manual* input.

        :param str question: Question to ask the user
        :param str default: default answer
        :return: Choice of the user
        :rtype: bool

    """
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)
    while True:
        sys.stdout.write(question + prompt)
        # Python 2.7 `input` attempts to evaluate the input, while in 3+
        # it returns a string. Python 2.7 `raw_input` returns a str as desired.
        if sys.version >= "3.0":
            choice = input().lower()
        else:
            choice = raw_input().lower()

        if default is not None and choice == "":
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' " "(or 'y' or 'n').\n")


def get_terminal(text="Password", confirm=False, allowedempty=False):
    import getpass

    while True:
        pw = getpass.getpass(text)
        if not pw and not allowedempty:
            print("Cannot be empty!")
            continue
        else:
            if not confirm:
                break
            pwck = getpass.getpass("Confirm " + text)
            if pw == pwck:
                break
            else:
                print("Not matching!")
    return pw


def format_operation_details(op, memos=False):
    if op[0] == "vote":
        return "%s: %s" % (
            op[1]["voter"],
            construct_identifier(op[1]["author"], op[1]["permlink"]),
        )
    elif op[0] == "comment":
        return "%s: %s" % (
            op[1]["author"],
            construct_identifier(op[1]["author"], op[1]["permlink"]),
        )
    elif op[0] == "transfer":
        str_ = "%s -> %s %s" % (op[1]["from"], op[1]["to"], op[1]["amount"],)

        if memos:
            memo = op[1]["memo"]
            if len(memo) > 0 and memo[0] == "#":
                hive = shared_hived_instance()
                # memo = hive.decode_memo(memo, op[1]["from"])
                memo = hive.decode_memo(memo, op)
            str_ += " (%s)" % memo
        return str_
    elif op[0] == "interest":
        return "%s" % (op[1]["interest"])
    else:
        return json.dumps(op[1], indent=4)


def print_permissions(account):
    t = PrettyTable(["Permission", "Threshold", "Key/Account"], hrules=0)
    t.align = "r"
    for permission in ["owner", "active", "posting"]:
        auths = []
        for type_ in ["account_auths", "key_auths"]:
            for authority in account[permission][type_]:
                auths.append("%s (%d)" % (authority[0], authority[1]))
        t.add_row(
            [permission, account[permission]["weight_threshold"], "\n".join(auths),]
        )
    print(t)


def print_json(tx):
    if sys.stdout.isatty():
        print(json.dumps(tx, indent=4))
    else:
        # You're being piped or redirected
        print(tx)


# this is another console script entrypoint
# also this function sucks and should be taken out back and shot
def hivetailentry():
    parser = argparse.ArgumentParser(
        description="UNIX tail(1)-like tool for the hive blockchain"
    )
    parser.add_argument(
        "-f", "--follow", help="Constantly stream output to stdout", action="store_true"
    )
    parser.add_argument(
        "-n", "--lines", type=int, default=10, help="How many ops to show"
    )
    parser.add_argument(
        "-j",
        "--json",
        help="Output as JSON instead of human-readable pretty-printed format",
        action="store_true",
    )
    args = parser.parse_args(sys.argv[1:])

    # FIXME(sneak) this always returns
    # '0000000000000000000000000000000000000000' for trx_id :(

    op_count = 0
    if args.json:
        if not args.follow:
            sys.stdout.write("[")
    for op in Blockchain().reliable_stream():
        if args.json:
            sys.stdout.write("%s" % json.dumps(op))
            if args.follow:
                sys.stdout.write("\n")  # for human eyeballs
                sys.stdout.flush()  # flush after each op if live mode
        else:
            pprint.pprint(op)
        op_count += 1
        if not args.follow:
            if op_count > args.lines:
                if args.json:
                    sys.stdout.write("]")
                return
            else:
                if args.json:
                    sys.stdout.write(",")



import json
import logging
import random
import re
import voluptuous as vo
from datetime import datetime, timedelta
from funcy.colls import none
from funcy.flow import silent
from funcy.seqs import first
from hivebase import memo
from hivebase import operations
from hivebase.account import PrivateKey, PublicKey
from hivebase.exceptions import AccountExistsException, MissingKeyError
from hivebase.storage import configStorage
from .account import Account
from .amount import Amount
from .converter import Converter
from .instance import shared_hived_instance
from .transactionbuilder import TransactionBuilder
from .utils import (
    derive_permlink,
    fmt_time_string,
    keep_in_dict,
    resolve_identifier,
)
from .wallet import Wallet

log = logging.getLogger(__name__)

HIVEIT_100_PERCENT = 10000
HIVEIT_1_PERCENT = HIVEIT_100_PERCENT / 100


# TODO
# account_witness_proxy [active]
# account_update [owner, active]
# decline_voting_rights [owner]

# challenge_authority [posting]
# prove_authority [active, owner]
# request_account_recovery [active]
# change_recovery_account [owner]
# reset_account [active]
# set_reset_account [active, owner]

# escrow_transfer [active]
# escrow_dispute [active]
# escrow_release [active]
# escrow_approve [active]
# custom_binary [any]

# custom [active]
# delete_comment [posting]

# enable_content_editing_operation [active]


class Commit(object):
    """Commit things to the Hive network.

    This class contains helper methods to construct, sign and broadcast
    common transactions, such as posting, voting, sending funds, etc.

    :param Hived hived: Hived node to connect to*
    :param bool offline: Do **not** broadcast transactions! *(optional)*
    :param bool debug: Enable Debugging *(optional)*

    :param list,dict,string keys: Predefine the wif keys to shortcut the
    wallet database

    Three wallet operation modes are possible:

    * **Wallet Database**: Here, the hivelibs load the keys from the
      locally stored wallet SQLite database (see ``storage.py``).
      To use this mode, simply call ``Hive()`` without the
      ``keys`` parameter
    * **Providing Keys**: Here, you can provide the keys for
      your accounts manually. All you need to do is add the wif
      keys for the accounts you want to use as a simple array
      using the ``keys`` parameter to ``Hive()``.
    * **Force keys**: This more is for advanced users and
      requires that you know what you are doing. Here, the
      ``keys`` parameter is a dictionary that overwrite the
      ``active``, ``owner``, ``posting`` or ``memo`` keys for
      any account. This mode is only used for *foreign*
      signatures!

    """

    def __init__(self, hived_instance=None, no_broadcast=False, **kwargs):
        self.hived = hived_instance or shared_hived_instance()
        self.no_broadcast = no_broadcast
        self.unsigned = kwargs.get("unsigned", False)
        self.expiration = int(kwargs.get("expiration", 30))

        self.wallet = Wallet(self.hived, **kwargs)

    def finalizeOp(self, ops, account, permission):
        """This method obtains the required private keys if present in
        the wallet, finalizes the transaction, signs it and
        broadacasts it

        :param operation ops: The operation (or list of operaions) to
            broadcast

        :param operation account: The account that authorizes the
            operation
        :param string permission: The required permission for
            signing (active, owner, posting)

        ... note::

            If ``ops`` is a list of operation, they all need to be
            signable by the same key! Thus, you cannot combine ops
            that require active permission with ops that require
            posting permission. Neither can you use different
            accounts for different operations!
        """
        tx = TransactionBuilder(
            None,
            hived_instance=self.hived,
            wallet_instance=self.wallet,
            no_broadcast=self.no_broadcast,
            expiration=self.expiration,
        )
        tx.appendOps(ops)
        if self.unsigned:
            tx.addSigningInformation(account, permission)
            return tx
        else:
            tx.appendSigner(account, permission)
            tx.sign()

        return tx.broadcast()

    def sign(self, unsigned_trx, wifs=[]):
        """Sign a provided transaction with the provided key(s)

        :param dict unsigned_trx: The transaction to be signed and returned
        :param string wifs: One or many wif keys to use for signing
            a transaction. If not present, the keys will be loaded
            from the wallet as defined in "missing_signatures" key
            of the transactions.
        """
        tx = TransactionBuilder(
            unsigned_trx,
            hived_instance=self.hived,
            wallet_instance=self.wallet,
            no_broadcast=self.no_broadcast,
            expiration=self.expiration,
        )
        tx.appendMissingSignatures(wifs)
        tx.sign()
        return tx.json()

    def broadcast(self, signed_trx):
        """Broadcast a transaction to the Hive network

        :param tx signed_trx: Signed transaction to broadcast
        """
        tx = TransactionBuilder(
            signed_trx,
            hived_instance=self.hived,
            wallet_instance=self.wallet,
            no_broadcast=self.no_broadcast,
            expiration=self.expiration,
        )
        return tx.broadcast()

    def post(
        self,
        title,
        body,
        author,
        permlink=None,
        reply_identifier=None,
        json_metadata=None,
        comment_options=None,
        community=None,
        tags=None,
        beneficiaries=None,
        self_vote=False,
    ):
        """Create a new post.

        If this post is intended as a reply/comment, `reply_identifier` needs
        to be set with the identifier of the parent post/comment (eg.
        `author/permlink`).

        Optionally you can also set json_metadata, comment_options and upvote
        the newly created post as an author.

        Setting category, tags or community will override the values provided
        in json_metadata and/or comment_options where appropriate.

        Args:

        title (str): Title of the post
        body (str): Body of the post/comment
        author (str): Account are you posting from
        permlink (str): Manually set the permlink (defaults to None).
            If left empty, it will be derived from title automatically.
        reply_identifier (str): Identifier of the parent post/comment (only
            if this post is a reply/comment).
        json_metadata (str, dict): JSON meta object that can be attached to
            the post.
        comment_options (str, dict): JSON options object that can be
            attached to the post.

        Example::
            comment_options = {
                'max_accepted_payout': '1000000.000 HBD',
                'percent_hive_dollars': 10000,
                'allow_votes': True,
                'allow_curation_rewards': True,
                'extensions': [[0, {
                    'beneficiaries': [
                        {'account': 'account1', 'weight': 5000},
                        {'account': 'account2', 'weight': 5000},
                    ]}
                ]]
            }

        community (str): (Optional) Name of the community we are posting
            into. This will also override the community specified in
            `json_metadata`.

        tags (str, list): (Optional) A list of tags (5 max) to go with the
            post. This will also override the tags specified in
            `json_metadata`. The first tag will be used as a 'category'. If
            provided as a string, it should be space separated.

        beneficiaries (list of dicts): (Optional) A list of beneficiaries
            for posting reward distribution. This argument overrides
            beneficiaries as specified in `comment_options`.

        For example, if we would like to split rewards between account1 and
        account2::

            beneficiaries = [
                {'account': 'account1', 'weight': 5000},
                {'account': 'account2', 'weight': 5000}
            ]

        self_vote (bool): (Optional) Upvote the post as author, right after
            posting.

        """

        # prepare json_metadata
        json_metadata = json_metadata or {}
        if isinstance(json_metadata, str):
            json_metadata = silent(json.loads)(json_metadata) or {}

        # override the community
        if community:
            json_metadata.update({"community": community})

        # deal with the category and tags
        if isinstance(tags, str):
            tags = list(set(filter(None, (re.split("[\W_]", tags)))))

        category = None
        tags = tags or json_metadata.get("tags", [])
        if tags:
            if len(tags) > 5:
                raise ValueError("Can only specify up to 5 tags per post.")

            # first tag should be a category
            category = tags[0]
            json_metadata.update({"tags": tags})

        # can't provide a category while replying to a post
        if reply_identifier and category:
            category = None

        # deal with replies/categories
        if reply_identifier:
            parent_author, parent_permlink = resolve_identifier(reply_identifier)
            if not permlink:
                permlink = derive_permlink(title, parent_permlink)
        elif category:
            parent_permlink = derive_permlink(category)
            parent_author = ""
            if not permlink:
                permlink = derive_permlink(title)
        else:
            parent_author = ""
            parent_permlink = ""
            if not permlink:
                permlink = derive_permlink(title)

        post_op = operations.Comment(
            **{
                "parent_author": parent_author,
                "parent_permlink": parent_permlink,
                "author": author,
                "permlink": permlink,
                "title": title,
                "body": body,
                "json_metadata": json_metadata,
            }
        )
        ops = [post_op]

        # if comment_options are used, add a new op to the transaction
        if comment_options or beneficiaries:
            options = keep_in_dict(
                comment_options or {},
                [
                    "max_accepted_payout",
                    "percent_hive_dollars",
                    "allow_votes",
                    "allow_curation_rewards",
                    "extensions",
                ],
            )
            # override beneficiaries extension
            if beneficiaries:
                # validate schema
                # or just simply vo.Schema([{'account': str, 'weight': int}])
                schema = vo.Schema(
                    [
                        {
                            vo.Required("account"): vo.All(str, vo.Length(max=16)),
                            vo.Required("weight", default=10000): vo.All(
                                int, vo.Range(min=1, max=10000)
                            ),
                        }
                    ]
                )
                schema(beneficiaries)
                options["beneficiaries"] = beneficiaries

            default_max_payout = "1000000.000 HBD"
            comment_op = operations.CommentOptions(
                **{
                    "author": author,
                    "permlink": permlink,
                    "max_accepted_payout": options.get(
                        "max_accepted_payout", default_max_payout
                    ),
                    "percent_hive_dollars": int(
                        options.get("percent_hive_dollars", 10000)
                    ),
                    "allow_votes": options.get("allow_votes", True),
                    "allow_curation_rewards": options.get(
                        "allow_curation_rewards", True
                    ),
                    "extensions": options.get("extensions", []),
                    "beneficiaries": options.get("beneficiaries"),
                }
            )
            ops.append(comment_op)

        if self_vote:
            vote_op = operations.Vote(
                **{
                    "voter": author,
                    "author": author,
                    "permlink": permlink,
                    "weight": 10000,
                }
            )
            ops.append(vote_op)

        return self.finalizeOp(ops, author, "posting")

    def vote(self, identifier, weight, account=None):
        """Vote for a post

        :param str identifier: Identifier for the post to upvote Takes
                               the form ``author/permlink``
        :param float weight: Voting weight. Range: -100.0 - +100.0. May
                             not be 0.0
        :param str account: Voter to use for voting. (Optional)

        If ``voter`` is not defines, the ``default_account`` will be taken
        or a ValueError will be raised

        .. code-block:: python

            hivepy set default_account <account>
        """
        if not account:
            account = configStorage.get("default_account")
        if not account:
            raise ValueError("You need to provide a voter account")

        post_author, post_permlink = resolve_identifier(identifier)

        op = operations.Vote(
            **{
                "voter": account,
                "author": post_author,
                "permlink": post_permlink,
                "weight": int(weight * HIVEIT_1_PERCENT),
            }
        )

        return self.finalizeOp(op, account, "posting")

    def create_account(
        self,
        account_name,
        json_meta=None,
        password=None,
        owner_key=None,
        active_key=None,
        posting_key=None,
        memo_key=None,
        additional_owner_keys=[],
        additional_active_keys=[],
        additional_posting_keys=[],
        additional_owner_accounts=[],
        additional_active_accounts=[],
        additional_posting_accounts=[],
        store_keys=True,
        store_owner_key=False,
        delegation_fee_hive="0 SHTO",
        creator=None,
    ):
        """Create new account in Hive

            The brainkey/password can be used to recover all generated keys
            (see `hivebase.account` for more details.

            By default, this call will use ``default_account`` to
            register a new name ``account_name`` with all keys being
            derived from a new brain key that will be returned. The
            corresponding keys will automatically be installed in the
            wallet.

            .. note:: Account creations cost a fee that is defined by
               the network. If you create an account, you will
               need to pay for that fee!

               **You can partially pay that fee by delegating VESTS.**

               To pay the fee in full in HIVE, leave
               ``delegation_fee_hive`` set to ``0 HIVE`` (Default).

               To pay the fee partially in HIVE, partially with delegated
               VESTS, set ``delegation_fee_hive`` to a value greater than ``1
               HIVE``. `Required VESTS will be calculated automatically.`

               To pay the fee with maximum amount of delegation, set
               ``delegation_fee_hive`` to ``1 HIVE``. `Required VESTS will be
               calculated automatically.`

            .. warning:: Don't call this method unless you know what
                          you are doing! Be sure to understand what this
                          method does and where to find the private keys
                          for your account.

        .. note:: Please note that this imports private keys (if password is
        present) into the wallet by default. However, it **does not import
        the owner key** unless `store_owner_key` is set to True (default
        False). Do NOT expect to be able to recover it from the wallet if
        you lose your password!

            :param str account_name: (**required**) new account name
            :param str json_meta: Optional meta data for the account
            :param str owner_key: Main owner key
            :param str active_key: Main active key
            :param str posting_key: Main posting key
            :param str memo_key: Main memo_key
            :param str password: Alternatively to providing keys, one
                                 can provide a password from which the
                                 keys will be derived
            :param list additional_owner_keys:  Additional owner public keys

            :param list additional_active_keys: Additional active public keys

            :param list additional_posting_keys: Additional posting public keys

            :param list additional_owner_accounts: Additional owner account
            names

            :param list additional_active_accounts: Additional active account
            names

            :param list additional_posting_accounts: Additional posting account
            names

            :param bool store_keys: Store new keys in the wallet (default:
            ``True``)

            :param bool store_owner_key: Store owner key in the wallet
            (default: ``False``)

            :param str delegation_fee_hive: (Optional) If set, `creator` pay a
            fee of this amount, and delegate the rest with VESTS (calculated
            automatically). Minimum: 1 HIVE. If left to 0 (Default), full fee
            is paid without VESTS delegation.

            :param str creator: which account should pay the registration fee
                                (defaults to ``default_account``)

        :raises AccountExistsException: if the account already exists on the
        blockchain

        """
        assert len(account_name) <= 16, "Account name must be at most 16 chars long"

        if not creator:
            creator = configStorage.get("default_account")
        if not creator:
            raise ValueError(
                "Not creator account given. Define it with "
                + "creator=x, or set the default_account using hivepy"
            )
        if password and (owner_key or posting_key or active_key or memo_key):
            raise ValueError("You cannot use 'password' AND provide keys!")

        # check if account already exists
        try:
            Account(account_name, hived_instance=self.hived)
        except:  # noqa FIXME(sneak)
            pass
        else:
            raise AccountExistsException

        " Generate new keys from password"
        from hivebase.account import PasswordKey, PublicKey

        if password:
            posting_key = PasswordKey(account_name, password, role="posting")
            active_key = PasswordKey(account_name, password, role="active")
            owner_key = PasswordKey(account_name, password, role="owner")
            memo_key = PasswordKey(account_name, password, role="memo")
            posting_pubkey = posting_key.get_public_key()
            active_pubkey = active_key.get_public_key()
            owner_pubkey = owner_key.get_public_key()
            memo_pubkey = memo_key.get_public_key()
            posting_privkey = posting_key.get_private_key()
            active_privkey = active_key.get_private_key()
            owner_privkey = owner_key.get_private_key()
            memo_privkey = memo_key.get_private_key()
            # store private keys
            if store_keys:
                if store_owner_key:
                    self.wallet.addPrivateKey(owner_privkey)
                self.wallet.addPrivateKey(active_privkey)
                self.wallet.addPrivateKey(posting_privkey)
                self.wallet.addPrivateKey(memo_privkey)
        elif owner_key and posting_key and active_key and memo_key:
            posting_pubkey = PublicKey(
                posting_key, prefix=self.hived.chain_params["prefix"]
            )
            active_pubkey = PublicKey(
                active_key, prefix=self.hived.chain_params["prefix"]
            )
            owner_pubkey = PublicKey(
                owner_key, prefix=self.hived.chain_params["prefix"]
            )
            memo_pubkey = PublicKey(memo_key, prefix=self.hived.chain_params["prefix"])
        else:
            raise ValueError(
                "Call incomplete! Provide either a password or public keys!"
            )

        owner = format(owner_pubkey, self.hived.chain_params["prefix"])
        active = format(active_pubkey, self.hived.chain_params["prefix"])
        posting = format(posting_pubkey, self.hived.chain_params["prefix"])
        memo = format(memo_pubkey, self.hived.chain_params["prefix"])

        owner_key_authority = [[owner, 1]]
        active_key_authority = [[active, 1]]
        posting_key_authority = [[posting, 1]]
        owner_accounts_authority = []
        active_accounts_authority = []
        posting_accounts_authority = []

        # additional authorities
        for k in additional_owner_keys:
            owner_key_authority.append([k, 1])
        for k in additional_active_keys:
            active_key_authority.append([k, 1])
        for k in additional_posting_keys:
            posting_key_authority.append([k, 1])

        for k in additional_owner_accounts:
            owner_accounts_authority.append([k, 1])
        for k in additional_active_accounts:
            active_accounts_authority.append([k, 1])
        for k in additional_posting_accounts:
            posting_accounts_authority.append([k, 1])

        props = self.hived.get_chain_properties()
        required_fee_hive = Amount(props["account_creation_fee"])

        required_fee_vests = 0
        delegation_fee_hive = Amount(delegation_fee_hive)
        if float(delegation_fee_hive):
            # creating accounts without delegation requires 30x
            # account_creation_fee creating account with delegation allows one
            # to use VESTS to pay the fee where the ratio must satisfy 1 HIVE
            # in fee == 5 HIVE in delegated VESTS

            delegated_hp_fee_mult = 5

            if float(delegation_fee_hive) < 1:
                raise ValueError(
                    "When creating account with delegation, at least "
                    + "1 HIVE in fee must be paid."
                )

            # calculate required remaining fee in vests
            remaining_fee = required_fee_hive - delegation_fee_hive
            if float(remaining_fee) > 0:
                required_hp = remaining_fee * delegated_hp_fee_mult
                required_fee_vests = Converter().hp_to_vests(required_hp) + 1

        s = {
            "creator": creator,
            "fee": "%s STEEM" % float(required_fee_hive),
            # "delegation": "%s VESTS" % required_fee_vests,
            "json_metadata": json_meta or {},
            "memo_key": memo,
            "new_account_name": account_name,
            "owner": {
                "account_auths": owner_accounts_authority,
                "key_auths": owner_key_authority,
                "weight_threshold": 1,
            },
            "active": {
                "account_auths": active_accounts_authority,
                "key_auths": active_key_authority,
                "weight_threshold": 1,
            },
            "posting": {
                "account_auths": posting_accounts_authority,
                "key_auths": posting_key_authority,
                "weight_threshold": 1,
            },
            "prefix": self.hived.chain_params["prefix"],
        }
        op = operations.AccountCreate(**s)

        return self.finalizeOp(op, creator, "active")

    def transfer(self, to, amount, asset, memo="", account=None):
        """Transfer HBD or HIVE to another account.

        :param str to: Recipient

        :param float amount: Amount to transfer

        :param str asset: Asset to transfer (``HBD`` or ``HIVE``)

        :param str memo: (optional) Memo, may begin with `#` for encrypted
        messaging

        :param str account: (optional) the source account for the transfer
        if not ``default_account``

        """
        if not account:
            account = configStorage.get("default_account")
        if not account:
            raise ValueError("You need to provide an account")

        # temporarily allow both steem and hive symbols until after HF24
        legacy_symbols = 0
        a = Account(account, hived_instance=self.hived)
        if "reward_steem_balance" in a:
            legacy_symbols = 1
        if legacy_symbols == 1:
            assert asset in ["HIVE", "HBD", "STEEM", "SBD"]
            if asset == "HIVE":
                asset = "STEEM"
            if asset == "HBD":
                asset = "SBD"

        if memo and memo[0] == "#":
            from hivebase import memo as Memo

            memo_wif = self.wallet.getMemoKeyForAccount(account)
            if not memo_wif:
                raise MissingKeyError("Memo key for %s missing!" % account)
            to_account = Account(to, hived_instance=self.hived)
            nonce = random.getrandbits(64)
            memo = Memo.encode_memo(
                PrivateKey(memo_wif),
                PublicKey(
                    to_account["memo_key"], prefix=self.hived.chain_params["prefix"]
                ),
                nonce,
                memo,
                prefix=self.hived.chain_params["prefix"],
            )

        op = operations.Transfer(
            **{
                "from": account,
                "to": to,
                "amount": "{:.{prec}f} {asset}".format(
                    float(amount), prec=3, asset=asset
                ),
                "memo": memo,
            }
        )
        print(op)
        return self.finalizeOp(op, account, "active")

    def withdraw_vesting(self, amount, account=None):
        """Withdraw VESTS from the vesting account.

        :param float amount: number of VESTS to withdraw over a period of
        104 weeks

        :param str account: (optional) the source account for the transfer
        if not ``default_account``

        """
        if not account:
            account = configStorage.get("default_account")
        if not account:
            raise ValueError("You need to provide an account")

        op = operations.WithdrawVesting(
            **{
                "account": account,
                "vesting_shares": "{:.{prec}f} {asset}".format(
                    float(amount), prec=6, asset="VESTS"
                ),
            }
        )

        return self.finalizeOp(op, account, "active")

    def transfer_to_vesting(self, amount, to=None, account=None):
        """Vest HIVE

        :param float amount: number of HIVE to vest

        :param str to: (optional) the source account for the transfer if not
        ``default_account``

        :param str account: (optional) the source account for the transfer
        if not ``default_account``

        """
        if not account:
            account = configStorage.get("default_account")
        if not account:
            raise ValueError("You need to provide an account")

        if not to:
            to = account  # powerup on the same account

        # temporarily allow both steem and hive symbols until after HF24
        asset_symbol = "HIVE"
        a = Account(account, hived_instance=self.hived)
        if "reward_steem_balance" in a:
            asset_symbol = "STEEM"

        op = operations.TransferToVesting(
            **{
                "from": account,
                "to": to,
                "amount": "{:.{prec}f} {asset}".format(
                    float(amount), prec=3, asset=asset_symbol
                ),
            }
        )

        return self.finalizeOp(op, account, "active")

    def convert(self, amount, account=None, request_id=None):
        """Convert HBD to HIVE (takes 3.5 days to settle)

        :param float amount: number of HBD to convert

        :param str account: (optional) the source account for the transfer
        if not ``default_account``

        :param str request_id: (optional) identifier for tracking the
        conversion`

        """
        if not account:
            account = configStorage.get("default_account")
        if not account:
            raise ValueError("You need to provide an account")

        if request_id:
            request_id = int(request_id)
        else:
            request_id = random.getrandbits(32)
        op = operations.Convert(
            **{
                "owner": account,
                "requestid": request_id,
                "amount": "{:.{prec}f} {asset}".format(
                    float(amount), prec=3, asset="HBD"
                ),
            }
        )

        return self.finalizeOp(op, account, "active")

    def claim_reward_balance(
        self,
        reward_hive="0 HIVE",
        reward_hbd="0 HBD",
        reward_vests="0 VESTS",
        account=None,
    ):
        """Claim reward balances.

        By default, this will claim ``all`` outstanding balances. To bypass
        this behaviour, set desired claim amount by setting any of
        `reward_hive`, `reward_hbd` or `reward_vests`.

        Args:
            reward_hive (string): Amount of HIVE you would like to claim.
            reward_hbd (string): Amount of HBD you would like to claim.
            reward_vests (string): Amount of VESTS you would like to claim.
            account (string): The source account for the claim if not
            ``default_account`` is used.
        """
        if not account:
            account = configStorage.get("default_account")
        if not account:
            raise ValueError("You need to provide an account")

        # if no values were set by user, claim all outstanding balances on
        # account
        if none(
            float(first(x.split(" "))) for x in [reward_hbd, reward_hive, reward_vests]
        ):
            a = Account(account, hived_instance=self.hived)
            # temporarily allow both steem and hive symbols until after HF24
            if "sbd_balance" in a:
                a["hbd_balance"] = a["sbd_balance"]
                a["savings_hbd_balance"] = a["savings_sbd_balance"]
                a["reward_hive_balance"] = a["reward_steem_balance"]
                a["reward_hbd_balance"] = a["reward_sbd_balance"]

            reward_hive = a["reward_hive_balance"]
            reward_hbd = a["reward_hbd_balance"]
            reward_vests = a["reward_vesting_balance"]

        # temporarily allow both steem and hive symbols until after HF24
        a = Account(account, hived_instance=self.hived)
        if "sbd_balance" in a:
            if reward_hive[-4:] == "HIVE":
                reward_hive = reward_hive[:-4] + "STEEM"
            if reward_hbd[-3:] == "HBD":
                reward_hbd = reward_hbd[:-3] + "SBD"

            op = operations.ClaimRewardBalance(
                **{
                    "account": account,
                    "reward_steem": reward_hive,
                    "reward_sbd": reward_hbd,
                    "reward_vests": reward_vests,
                }
            )
        else:
            op = operations.ClaimRewardBalance(
                **{
                    "account": account,
                    "reward_hive": reward_hive,
                    "reward_hbd": reward_hbd,
                    "reward_vests": reward_vests,
                }
            )
        return self.finalizeOp(op, account, "posting")

    def delegate_vesting_shares(self, to_account, vesting_shares, account=None):
        """Delegate HP to another account.

        Args:
            to_account (string): Account we are delegating shares to
            (delegatee).
            vesting_shares (string): Amount of VESTS to delegate eg. `10000
            VESTS`.
            account (string): The source account (delegator). If not specified,
            ``default_account`` is used.
        """
        if not account:
            account = configStorage.get("default_account")
        if not account:
            raise ValueError("You need to provide an account")

        op = operations.DelegateVestingShares(
            **{
                "delegator": account,
                "delegatee": to_account,
                "vesting_shares": str(Amount(vesting_shares)),
            }
        )
        return self.finalizeOp(op, account, "active")

    def witness_feed_publish(self, hive_usd_price, quote="1.000", account=None):
        """Publish a feed price as a witness.

        :param float hive_usd_price: Price of HIVE in USD (implied price)
        :param float quote: (optional) Quote Price. Should be 1.000, unless
        we are adjusting the feed to support the peg.
        :param str account: (optional) the source account for the transfer
        if not ``default_account``
        """
        if not account:
            account = configStorage.get("default_account")
        if not account:
            raise ValueError("You need to provide an account")

        op = operations.FeedPublish(
            **{
                "publisher": account,
                "exchange_rate": {
                    "base": "%s HBD" % hive_usd_price,
                    "quote": "%s HIVE" % quote,
                },
            }
        )
        return self.finalizeOp(op, account, "active")

    def witness_update(self, signing_key, url, props, account=None):
        """Update witness

        :param pubkey signing_key: Signing key
        :param str url: URL
        :param dict props: Properties
        :param str account: (optional) witness account name

         Properties:::

            {
                "account_creation_fee": x,
                "maximum_block_size": x,
                "hbd_interest_rate": x,
            }

        """
        if not account:
            account = configStorage.get("default_account")
        if not account:
            raise ValueError("You need to provide an account")

        try:
            PublicKey(signing_key)
        except Exception as e:
            raise e

        op = operations.WitnessUpdate(
            **{
                "owner": account,
                "url": url,
                "block_signing_key": signing_key,
                "props": props,
                "fee": "0.000 HIVE",
                "prefix": self.hived.chain_params["prefix"],
            }
        )
        return self.finalizeOp(op, account, "active")

    def witness_set_properties(self, signing_key, props, account=None):
        """Update witness

        :param pubkey signing_key: Signing key
        :param dict props: Properties
        :param str account: (optional) witness account name

         Properties:::

            [
                ["account_creation_fee": x]
                ["maximum_block_size": x]
                ["sbd_interest_rate": x]
            ]

        """
        if not account:
            account = configStorage.get("default_account")
        if not account:
            raise ValueError("You need to provide an account")

        # temporarily allow both steem and hive symbols until after HF24
        legacy_symbols = 0
        a = Account(account, hived_instance=self.hived)
        if "reward_steem_balance" in a:
            legacy_symbols = 1

        props_list = [
            [
                "key",
                repr(
                    PrivateKey(
                        signing_key, prefix=self.hived.chain_params["prefix"]
                    ).pubkey
                ),
            ]
        ]
        for k in props:
            if k == "account_creation_fee":
                # temporarily allow both steem and hive symbols until after HF24
                if props[k][-4:] == "HIVE" and legacy_symbols == 1:
                    props[k] = props[k][:-4] + "STEEM"
            props_list.append([k, props[k]])

        op = operations.WitnessSetProperties(
            **{"owner": account, "props": props_list, "extensions": []}
        )
        tb = TransactionBuilder(
            None,
            hived_instance=self.hived,
            wallet_instance=self.wallet,
            no_broadcast=self.no_broadcast,
            expiration=self.expiration,
        )
        tb.appendOps([op])
        tb.appendWif(signing_key)
        tb.sign()
        return tb.broadcast()

    def decode_memo(self, enc_memo):
        """Try to decode an encrypted memo"""
        assert enc_memo[0] == "#", "decode memo requires memos to start with '#'"
        keys = memo.involved_keys(enc_memo)
        wif = None
        for key in keys:
            wif = self.wallet.getPrivateKeyForPublicKey(str(key))
            if wif:
                break
        if not wif:
            raise MissingKeyError
        return memo.decode_memo(PrivateKey(wif), enc_memo)

    def interest(self, account):
        """Caluclate interest for an account

        :param str account: Account name to get interest for
        """
        account = Account(account, hived_instance=self.hived)
        last_payment = fmt_time_string(account["hbd_last_interest_payment"])
        next_payment = last_payment + timedelta(days=30)
        interest_rate = (
            self.hived.get_dynamic_global_properties()["hbd_interest_rate"] / 100
        )  # percent
        interest_amount = (
            (interest_rate / 100)
            * int(int(account["hbd_seconds"]) / (60 * 60 * 24 * 356))
            * 10 ** -3
        )

        return {
            "interest": interest_amount,
            "last_payment": last_payment,
            "next_payment": next_payment,
            "next_payment_duration": next_payment - datetime.now(),
            "interest_rate": interest_rate,
        }

    def set_withdraw_vesting_route(
        self, to, percentage=100, account=None, auto_vest=False
    ):
        """Set up a vesting withdraw route. When vesting shares are
        withdrawn, they will be routed to these accounts based on the
        specified weights.

        :param str to: Recipient of the vesting withdrawal
        :param float percentage: The percent of the withdraw to go
            to the 'to' account.
        :param str account: (optional) the vesting account
        :param bool auto_vest: Set to true if the from account
            should receive the VESTS as VESTS, or false if it should
            receive them as HIVE. (defaults to ``False``)
        """
        if not account:
            account = configStorage.get("default_account")
        if not account:
            raise ValueError("You need to provide an account")

        op = operations.SetWithdrawVestingRoute(
            **{
                "from_account": account,
                "to_account": to,
                "percent": int(percentage * HIVEIT_1_PERCENT),
                "auto_vest": auto_vest,
            }
        )

        return self.finalizeOp(op, account, "active")

    @staticmethod
    def _test_weights_treshold(authority):
        weights = 0
        for a in authority["account_auths"]:
            weights += a[1]
        for a in authority["key_auths"]:
            weights += a[1]
        if authority["weight_threshold"] > weights:
            raise ValueError("Threshold too restrictive!")

    def allow(
        self, foreign, weight=None, permission="posting", account=None, threshold=None
    ):
        """Give additional access to an account by some other public
        key or account.

        :param str foreign: The foreign account that will obtain access
        :param int weight: (optional) The weight to use. If not
            define, the threshold will be used. If the weight is
            smaller than the threshold, additional signatures will
            be required. (defaults to threshold)
        :param str permission: (optional) The actual permission to
            modify (defaults to ``posting``)
        :param str account: (optional) the account to allow access
            to (defaults to ``default_account``)
        :param int threshold: The threshold that needs to be reached
            by signatures to be able to interact
        """
        if not account:
            account = configStorage.get("default_account")
        if not account:
            raise ValueError("You need to provide an account")

        if permission not in ["owner", "posting", "active"]:
            raise ValueError(
                "Permission needs to be either 'owner', 'posting', or 'active"
            )
        account = Account(account, hived_instance=self.hived)
        if not weight:
            weight = account[permission]["weight_threshold"]

        authority = account[permission]
        try:
            pubkey = PublicKey(foreign)
            authority["key_auths"].append([str(pubkey), weight])
        except:  # noqa FIXME(sneak)
            try:
                foreign_account = Account(foreign, hived_instance=self.hived)
                authority["account_auths"].append([foreign_account["name"], weight])
            except:  # noqa FIXME(sneak)
                raise ValueError("Unknown foreign account or unvalid public key")
        if threshold:
            authority["weight_threshold"] = threshold
            self._test_weights_treshold(authority)

        op = operations.AccountUpdate(
            **{
                "account": account["name"],
                permission: authority,
                "memo_key": account["memo_key"],
                "json_metadata": account["json_metadata"],
                "prefix": self.hived.chain_params["prefix"],
            }
        )
        if permission == "owner":
            return self.finalizeOp(op, account["name"], "owner")
        else:
            return self.finalizeOp(op, account["name"], "active")

    def disallow(self, foreign, permission="posting", account=None, threshold=None):
        """Remove additional access to an account by some other public
        key or account.

        :param str foreign: The foreign account that will obtain access
        :param str permission: (optional) The actual permission to
            modify (defaults to ``posting``)
        :param str account: (optional) the account to allow access
            to (defaults to ``default_account``)
        :param int threshold: The threshold that needs to be reached
            by signatures to be able to interact
        """
        if not account:
            account = configStorage.get("default_account")
        if not account:
            raise ValueError("You need to provide an account")

        if permission not in ["owner", "posting", "active"]:
            raise ValueError(
                "Permission needs to be either 'owner', 'posting', or 'active"
            )
        account = Account(account, hived_instance=self.hived)
        authority = account[permission]

        try:
            pubkey = PublicKey(foreign, prefix=self.hived.chain_params["prefix"])
            affected_items = list(
                filter(lambda x: x[0] == str(pubkey), authority["key_auths"])
            )
            authority["key_auths"] = list(
                filter(lambda x: x[0] != str(pubkey), authority["key_auths"])
            )
        except:  # noqa FIXME(sneak)
            try:
                foreign_account = Account(foreign, hived_instance=self.hived)
                affected_items = list(
                    filter(
                        lambda x: x[0] == foreign_account["name"],
                        authority["account_auths"],
                    )
                )
                authority["account_auths"] = list(
                    filter(
                        lambda x: x[0] != foreign_account["name"],
                        authority["account_auths"],
                    )
                )
            except:  # noqa FIXME(sneak)
                raise ValueError("Unknown foreign account or unvalid public key")

        removed_weight = affected_items[0][1]

        # Define threshold
        if threshold:
            authority["weight_threshold"] = threshold

        # Correct threshold (at most by the amount removed from the
        # authority)
        try:
            self._test_weights_treshold(authority)
        except:  # noqa FIXME(sneak)
            log.critical(
                "The account's threshold will be reduced by %d" % removed_weight
            )
            authority["weight_threshold"] -= removed_weight
            self._test_weights_treshold(authority)

        op = operations.AccountUpdate(
            **{
                "account": account["name"],
                permission: authority,
                "memo_key": account["memo_key"],
                "json_metadata": account["json_metadata"],
            }
        )
        if permission == "owner":
            return self.finalizeOp(op, account["name"], "owner")
        else:
            return self.finalizeOp(op, account["name"], "active")

    def update_memo_key(self, key, account=None):
        """Update an account's memo public key

        This method does **not** add any private keys to your
        wallet but merely changes the memo public key.

        :param str key: New memo public key
        :param str account: (optional) the account to allow access
            to (defaults to ``default_account``)
        """
        if not account:
            account = configStorage.get("default_account")
        if not account:
            raise ValueError("You need to provide an account")

        PublicKey(key)  # raises exception if invalid
        account = Account(account, hived_instance=self.hived)
        op = operations.AccountUpdate(
            **{
                "account": account["name"],
                "memo_key": key,
                "json_metadata": account["json_metadata"],
            }
        )
        return self.finalizeOp(op, account["name"], "active")

    def approve_witness(self, witness, account=None, approve=True):
        """Vote **for** a witness. This method adds a witness to your
        set of approved witnesses. To remove witnesses see
        ``disapprove_witness``.

        :param str witness: witness to approve
        :param str account: (optional) the account to allow access
            to (defaults to ``default_account``)
        """
        if not account:
            account = configStorage.get("default_account")
        if not account:
            raise ValueError("You need to provide an account")
        account = Account(account, hived_instance=self.hived)
        op = operations.AccountWitnessVote(
            **{
                "account": account["name"],
                "witness": witness,
                "approve": approve,
            }
        )
        return self.finalizeOp(op, account["name"], "active")

    def disapprove_witness(self, witness, account=None):
        """Remove vote for a witness. This method removes
        a witness from your set of approved witnesses. To add
        witnesses see ``approve_witness``.

        :param str witness: witness to approve
        :param str account: (optional) the account to allow access
            to (defaults to ``default_account``)
        """
        return self.approve_witness(witness=witness, account=account, approve=False)

    def custom_json(self, id, json, required_auths=[], required_posting_auths=[]):
        """Create a custom json operation

        :param str id: identifier for the custom json (max length 32 bytes)
        :param json json: the json data to put into the custom_json
            operation
        :param list required_auths: (optional) required auths
        :param list required_posting_auths: (optional) posting auths
        """
        account = None
        if len(required_auths):
            account = required_auths[0]
        elif len(required_posting_auths):
            account = required_posting_auths[0]
        else:
            raise Exception("At least one account needs to be specified")
        op = operations.CustomJson(
            **{
                "json": json,
                "required_auths": required_auths,
                "required_posting_auths": required_posting_auths,
                "id": id,
            }
        )
        return self.finalizeOp(op, account, "posting")

    def reblog(self, identifier, account=None):
        """Reblog a post

        :param str identifier: post identifier (<account>/<permlink>)
        :param str account: (optional) the account to allow access
            to (defaults to ``default_account``)
        """
        if not account:
            account = configStorage.get("default_account")
        if not account:
            raise ValueError("You need to provide an account")
        author, permlink = resolve_identifier(identifier)
        json_body = [
            "reblog",
            {"account": account, "author": author, "permlink": permlink},
        ]
        return self.custom_json(
            id="follow", json=json_body, required_posting_auths=[account]
        )

    def unfollow(self, unfollow, what=["blog"], account=None):
        """Unfollow another account's blog

        :param str unfollow: Follow this account
        :param list what: List of states to follow
            (defaults to ``['blog']``)
        :param str account: (optional) the account to allow access
            to (defaults to ``default_account``)
        """
        # FIXME: removing 'blog' from the array requires to first read
        # the follow.what from the blockchain
        return self.follow(unfollow, what=[], account=account)

    def follow(self, follow, what=["blog"], account=None):
        """Follow another account's blog

        :param str follow: Follow this account
        :param list what: List of states to follow
            (defaults to ``['blog']``)
        :param str account: (optional) the account to allow access
            to (defaults to ``default_account``)
        """
        if not account:
            account = configStorage.get("default_account")
        if not account:
            raise ValueError("You need to provide an account")

        json_body = ["follow", {"follower": account, "following": follow, "what": what}]
        return self.custom_json(
            id="follow", json=json_body, required_posting_auths=[account]
        )

    def update_account_profile(self, profile, account=None):
        """Update an account's meta data (json_meta)

        :param dict json: The meta data to use (i.e. use Profile() from
            account.py)
        :param str account: (optional) the account to allow access
            to (defaults to ``default_account``)
        """
        if not account:
            account = configStorage.get("default_account")
        if not account:
            raise ValueError("You need to provide an account")
        account = Account(account, hived_instance=self.hived)
        op = operations.AccountUpdate(
            **{
                "account": account["name"],
                "memo_key": account["memo_key"],
                "json_metadata": profile,
            }
        )
        return self.finalizeOp(op, account["name"], "active")

    def comment_options(self, identifier, options, account=None):
        """Set the comment options

        :param str identifier: Post identifier
        :param dict options: The options to define.
        :param str account: (optional) the account to allow access
            to (defaults to ``default_account``)

        For the options, you have these defaults:::

                {
                    "author": "",
                    "permlink": "",
                    "max_accepted_payout": "1000000.000 HBD",
                    "percent_hive_dollars": 10000,
                    "allow_votes": True,
                    "allow_curation_rewards": True,
                }

        """
        if not account:
            account = configStorage.get("default_account")
        if not account:
            raise ValueError("You need to provide an account")
        account = Account(account, hived_instance=self.hived)
        author, permlink = resolve_identifier(identifier)
        default_max_payout = "1000000.000 HBD"
        op = operations.CommentOptions(
            **{
                "author": author,
                "permlink": permlink,
                "max_accepted_payout": options.get(
                    "max_accepted_payout", default_max_payout
                ),
                "percent_hive_dollars": options.get("percent_hive_dollars", 100)
                * HIVEIT_1_PERCENT,
                "allow_votes": options.get("allow_votes", True),
                "allow_curation_rewards": options.get("allow_curation_rewards", True),
            }
        )
        return self.finalizeOp(op, account["name"], "posting")


if __name__ == "__main__":
    pass



import random

from hivebase import transactions, operations
from hivebase.storage import configStorage as config

from .amount import Amount
from .instance import shared_hived_instance


class Dex(object):
    """ This class allows to access calls specific for the internal
        exchange of HIVE.

        :param Hived hived_instance: Hived() instance to use when
        accessing a RPC

    """
    assets = ["HIVE", "HBD", "STEEM", "SBD"]

    def __init__(self, hived_instance=None):
        self.hived = hived_instance or shared_hived_instance()

    def _get_asset(self, symbol):
        """ Return the properties of the assets tradeable on the
            network.

            :param str symbol: Symbol to get the data for (i.e. HIVE, HBD,
            VESTS)

        """
        # temporarily allow both steem and hive symbols until after HF24
        precision = 3
        if symbol == 'VESTS':
            precision = 6
        return {"symbol":symbol,"precision": precision}
#        if symbol == "HIVE" or symbol == "STEEM":
#            return {"symbol": "HIVE", "precision": 3}
#        elif symbol == "HBD" or symbol == "SBD":
#            return {"symbol": "HBD", "precision": 3}
#        elif symbol == "VESTS":
#            return {"symbol": "VESTS", "precision": 6}
#        else:
#            return None

    def _get_assets(self, quote):
        """ Given the `quote` asset, return base. If quote is HBD, then
            base is HIVE and vice versa.
        """
        assets = self.assets.copy()
        assets.remove(quote)
        base = assets[0]
        return self._get_asset(quote), self._get_asset(base)

    def get_ticker(self):
        """ Returns the ticker for all markets.

            Output Parameters:

            * ``latest``: Price of the order last filled
            * ``lowest_ask``: Price of the lowest ask
            * ``highest_bid``: Price of the highest bid
            * ``hbd_volume``: Volume of HBD
            * ``hive_volume``: Volume of HIVE
            * ``percent_change``: 24h change percentage (in %)

            .. note::

                Market is HIVE:HBD and prices are HBD per HIVE!

            Sample Output:

            .. code-block:: js

                 {'highest_bid': 0.30100226633322913,
                  'latest': 0.0,
                  'lowest_ask': 0.3249636958897082,
                  'percent_change': 0.0,
                  'hbd_volume': 108329611.0,
                  'hive_volume': 355094043.0}


        """
        t = self.hived.get_ticker()
        return {
            'highest_bid': float(t['highest_bid']),
            'latest': float(t["latest"]),
            'lowest_ask': float(t["lowest_ask"]),
            'percent_change': float(t["percent_change"]),
            'hbd_volume': Amount(t["hbd_volume"]),
            'hive_volume': Amount(t["hive_volume"])
        }

    def trade_history(self, time=1 * 60 * 60, limit=100):
        """ Returns the trade history for the internal market

            :param int time: Show the last x seconds of trades (default 1h)
            :param int limit: amount of trades to show (<100) (default: 100)
        """
        assert limit <= 100, "'limit' has to be smaller than 100"
        return self.hived.get_trade_history(
            transactions.fmt_time_from_now(-time),
            transactions.fmt_time_from_now(),
            limit,
        )

    def market_history_buckets(self):
        return self.hived.get_market_history_buckets()

    def market_history(
            self,
            bucket_seconds=60 * 5,
            start_age=1 * 60 * 60,
            end_age=0,
    ):
        """ Return the market history (filled orders).

            :param int bucket_seconds: Bucket size in seconds (see
            `returnMarketHistoryBuckets()`)

            :param int start_age: Age (in seconds) of the start of the
            window (default: 1h/3600)

            :param int end_age: Age (in seconds) of the end of the window
            (default: now/0)

            Example:

            .. code-block:: js

                 {'close_hbd': 2493387,
                  'close_hive': 7743431,
                  'high_hbd': 1943872,
                  'high_hive': 5999610,
                  'id': '7.1.5252',
                  'low_hbd': 534928,
                  'low_hive': 1661266,
                  'open': '2016-07-08T11:25:00',
                  'open_hbd': 534928,
                  'open_hive': 1661266,
                  'hbd_volume': 9714435,
                  'seconds': 300,
                  'hive_volume': 30088443},
        """
        return self.hived.get_market_history(
            bucket_seconds,
            transactions.fmt_time_from_now(-start_age - end_age),
            transactions.fmt_time_from_now(-end_age),
        )

    def buy(self,
            amount,
            quote_symbol,
            rate,
            expiration=7 * 24 * 60 * 60,
            killfill=False,
            account=None,
            order_id=None):
        """ Places a buy order in a given market (buy ``quote``, sell
            ``base`` in market ``quote_base``). If successful, the
            method will return the order creating (signed) transaction.

            :param number amount: Amount of ``quote`` to buy

            :param str quote_symbol: HIVE, or HBD

            :param float price: price denoted in ``base``/``quote``

            :param number expiration: (optional) expiration time of the
            order in seconds (defaults to 7 days)

            :param bool killfill: flag that indicates if the order shall be
            killed if it is not filled (defaults to False)

            :param str account: (optional) the source account for the
            transfer if not ``default_account``

            :param int order_id: (optional) a 32bit orderid for tracking of
            the created order (random by default)

            Prices/Rates are denoted in 'base', i.e. the HIVE:HBD market
            is priced in HBD per HIVE.
        """
        if not account:
            if "default_account" in config:
                account = config["default_account"]
        if not account:
            raise ValueError("You need to provide an account")

        # We buy quote and pay with base
        quote, base = self._get_assets(quote=quote_symbol)
        op = operations.LimitOrderCreate(
            **{
                "owner":
                    account,
                "orderid":
                    order_id or random.getrandbits(32),
                "amount_to_sell":
                    '{:.{prec}f} {asset}'.format(
                        amount * rate,
                        prec=base["precision"],
                        asset=base["symbol"]),
                "min_to_receive":
                    '{:.{prec}f} {asset}'.format(
                        amount, prec=quote["precision"], asset=quote["symbol"]),
                "fill_or_kill":
                    killfill,
                "expiration":
                    transactions.fmt_time_from_now(expiration)
            })
        return self.hived.commit.finalizeOp(op, account, "active")

    def sell(self,
             amount,
             quote_symbol,
             rate,
             expiration=7 * 24 * 60 * 60,
             killfill=False,
             account=None,
             orderid=None):
        """ Places a sell order in a given market (sell ``quote``, buy
            ``base`` in market ``quote_base``). If successful, the
            method will return the order creating (signed) transaction.

            :param number amount: Amount of ``quote`` to sell

            :param str quote_symbol: HIVE, or HBD

            :param float price: price denoted in ``base``/``quote``

            :param number expiration: (optional) expiration time of the
            order in seconds (defaults to 7 days)

            :param bool killfill: flag that indicates if the order shall be
            killed if it is not filled (defaults to False)

            :param str account: (optional) the source account for the
            transfer if not ``default_account``

            :param int orderid: (optional) a 32bit orderid for tracking of
            the created order (random by default)

            Prices/Rates are denoted in 'base', i.e. the HIVE:HBD market
            is priced in HBD per HIVE.
        """
        if not account:
            if "default_account" in config:
                account = config["default_account"]
        if not account:
            raise ValueError("You need to provide an account")
        # We buy quote and pay with base
        quote, base = self._get_assets(quote=quote_symbol)
        op = operations.LimitOrderCreate(
            **{
                "owner":
                    account,
                "orderid":
                    orderid or random.getrandbits(32),
                "amount_to_sell":
                    '{:.{prec}f} {asset}'.format(
                        amount, prec=quote["precision"], asset=quote["symbol"]),
                "min_to_receive":
                    '{:.{prec}f} {asset}'.format(
                        amount * rate,
                        prec=base["precision"],
                        asset=base["symbol"]),
                "fill_or_kill":
                    killfill,
                "expiration":
                    transactions.fmt_time_from_now(expiration)
            })
        return self.hived.commit.finalizeOp(op, account, "active")

    def cancel(self, orderid, account=None):
        """ Cancels an order you have placed in a given market.

            :param int orderid: the 32bit orderid

            :param str account: (optional) the source account for the
            transfer if not ``default_account``

        """
        if not account:
            if "default_account" in config:
                account = config["default_account"]
        if not account:
            raise ValueError("You need to provide an account")

        op = operations.LimitOrderCancel(**{
            "owner": account,
            "orderid": orderid,
        })
        return self.hived.commit.finalizeOp(op, account, "active")



import math

from .amount import Amount
from .instance import shared_hived_instance


class Converter(object):
    """ Converter simplifies the handling of different metrics of
        the blockchain

        :param Hived hived_instance: Hived() instance to
        use when accessing a RPC

    """

    def __init__(self, hived_instance=None):
        self.hived = hived_instance or shared_hived_instance()

        self.CONTENT_CONSTANT = 2000000000000

    def hbd_median_price(self):
        """ Obtain the hbd price as derived from the median over all
            witness feeds. Return value will be HBD
        """
        return (Amount(self.hived.get_feed_history()['current_median_history']
                       ['base']).amount / Amount(self.hived.get_feed_history(
        )['current_median_history']['quote']).amount)

    def hive_per_mvests(self):
        """ Obtain HIVE/MVESTS ratio
        """
        info = self.hived.get_dynamic_global_properties()

        # temporarily allow both steem and hive symbols until after HF24
        if 'total_vesting_fund_steem' in info:
            info['total_vesting_fund_hive'] = info['total_vesting_fund_steem']

        return (Amount(info["total_vesting_fund_hive"]).amount /
                (Amount(info["total_vesting_shares"]).amount / 1e6))

    def vests_to_hp(self, vests):
        """ Obtain HP from VESTS (not MVESTS!)

            :param number vests: Vests to convert to HP
        """
        return vests / 1e6 * self.hive_per_mvests()

    def hp_to_vests(self, hp):
        """ Obtain VESTS (not MVESTS!) from HP

            :param number sp: HP to convert
        """
        return hp * 1e6 / self.hive_per_mvests()

    def hp_to_rshares(self, hp, voting_power=10000, vote_pct=10000):
        """ Obtain the r-shares

            :param number hp: Hive Power
            :param int voting_power: voting power (100% = 10000)
            :param int vote_pct: voting participation (100% = 10000)
        """
        # calculate our account voting shares (from vests), mine is 6.08b
        vesting_shares = int(self.hp_to_vests(hp) * 1e6)

        # get props
        props = self.hived.get_dynamic_global_properties()

        # determine voting power used
        used_power = int((voting_power * vote_pct) / 10000);
        max_vote_denom = props['vote_power_reserve_rate'] * (5 * 60 * 60 * 24) / (60 * 60 * 24);
        used_power = int((used_power + max_vote_denom - 1) / max_vote_denom)

        # calculate vote rshares
        rshares = ((vesting_shares * used_power) / 10000)

        return rshares

    def hive_to_hbd(self, amount_hive):
        """ Conversion Ratio for given amount of HIVE to HBD at current
            price feed

            :param number hive: Amount of HIVE
        """
        return self.hbd_median_price() * amount_hive

    def hive(self, amount_hbd):
        """ Conversion Ratio for given amount of HBD to HIVE at current
            price feed

            :param number amount_hbd: Amount of HBD
        """
        return amount_hbd / self.hbd_median_price()

    def hbd_to_rshares(self, hbd_payout):
        """ Obtain r-shares from HBD

            :param number hbd_payout: Amount of HBD
        """
        hive_payout = self.hbd_to_hive(hbd_payout)

        reward_fund = self.hived.get_reward_fund()
        reward_balance = Amount(reward_fund['reward_balance']).amount
        recent_claims = int(reward_fund['recent_claims'])

        return int(recent_claims * hive_payout / (reward_balance - hive_payout))

    def rshares_2_weight(self, rshares):
        """ Obtain weight from rshares

            :param number rshares: R-Shares
        """
        _max = 2 ** 64 - 1
        return (_max * rshares) / (2 * self.CONTENT_CONSTANT + rshares)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command
from setuptools.command.test import test as TestCommand

# Package meta-data.
NAME = "shareberry"
VERSION = "0.0.5"
DESCRIPTION = "A python shareberry library."
URL = "https://gitlab.com"
EMAIL = "dev@shareberry.io"
AUTHOR = "pharesim"

# What packages are required for this module to be executed?
REQUIRED = [
    "appdirs",
    "certifi",
    "ecdsa>=0.13",
    "funcy",
    'futures ; python_version < "3.0.0"',
    "future",
    "langdetect",
    "prettytable",
    "pycrypto>=1.9.1",
    "pylibscrypt>=1.6.1",
    "scrypt>=0.8.0",
    "toolz",
    "ujson",
    "urllib3",
    "voluptuous",
    "w3lib",
]
TEST_REQUIRED = [
    "pep8",
    "pytest",
    'pytest-pylint ; python_version >= "3.4.0"',
    "pytest-xdist",
    "pytest-runner",
    "pytest-pep8",
    "pytest-cov",
    "yapf",
    "autopep8",
]

BUILD_REQUIRED = [
    "twine",
    "pypandoc",
    "recommonmark" "wheel",
    "setuptools",
    "sphinx",
    "sphinx_rtd_theme",
]
# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.rst' is present in your MANIFEST.in file!
# with io.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
#     long_description = '\n' + f.read()


class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass into py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        try:
            from multiprocessing import cpu_count

            self.pytest_args = ["-n", str(cpu_count()), "--boxed"]
        except (ImportError, NotImplementedError):
            self.pytest_args = ["-n", "1", "--boxed"]

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPi via Twine…")
        os.system("twine upload dist/*")

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    keywords=["hive", "cryptocurrency", "blockchain"],
    # long_description=long_description,
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    packages=find_packages(exclude=("tests", "scripts")),
    entry_points={
        "console_scripts": [
            "piston=hive.cli:legacyentry",
            "hivepy=hive.cli:legacyentry",
            "hivetail=hive.cli:hivetailentry",
        ],
    },
    install_requires=REQUIRED,
    extras_require={
        "dev": TEST_REQUIRED + BUILD_REQUIRED,
        "build": BUILD_REQUIRED,
        "test": TEST_REQUIRED,
    },
    tests_require=TEST_REQUIRED,
    include_package_data=True,
    license="MIT",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Development Status :: 4 - Beta",
    ],
    # $ setup.py publish support.
    cmdclass={"upload": UploadCommand, "test": PyTest},
)



import hive as hve
import sys

_shared_hived_instance = None


def get_config_node_list():
    from hivebase.storage import configStorage
    nodes = configStorage.get('nodes', None)
    if nodes:
        return nodes.split(',')


def shared_hived_instance():
    """ This method will initialize _shared_hived_instance and return it.
    The purpose of this method is to have offer single default Hive
    instance that can be reused by multiple classes.  """

    global _shared_hived_instance
    if not _shared_hived_instance:
        if sys.version >= '3.0':
            _shared_hived_instance = hve.hived.Hived(
                nodes=get_config_node_list())
        else:
            _shared_hived_instance = hve.Hived(
                nodes=get_config_node_list())
    return _shared_hived_instance


def set_shared_hived_instance(hived_instance):
    """ This method allows us to override default hive instance for all
    users of _shared_hived_instance.  """

    global _shared_hived_instance
    _shared_hived_instance = hived_instance
